from yta_editor.tracks.items.abstract import TrackItemWithAudioMedia
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


class _AudioTrackItem(TrackItemWithAudioMedia):
    """
    Class to represent an element that is on the
    track, that can be an empty space or an audio.
    """

    def __init__(
        self,
        track: 'AudioTrack',
        t_start: Union[int, float, Fraction],
        t_end: Union[int, float, Fraction],
        media: Union['AudioTimed', None] = None,
        item_in: Union['_AudioTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', 'GapTrackItem', None] = None
    ):
        ParameterValidator.validate_instance_of('media', media, 'AudioTimed')

        super().__init__(
            track = track,
            t_start = t_start,
            t_end = t_end,
            media = media,
            item_in = item_in,
            item_out = item_out
        )


from yta_numpy.rgba.utils import RGBAFrameUtils
from yta_editor_nodes.old_opengl.effects import AudioEffectsStack, VideoEffectsStack
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from quicktions import Fraction
from typing import Union


# TODO: These methods below are old version and
# now we are implementing effects through nodes
"""
These methods below are shared by the
Audio and Video class that handle and
wrap an audio or video.
"""
def apply_video_effects_to_frame_at(
    effects_stack: VideoEffectsStack,
    frame: VideoFrame,
    t: Union[int, float, 'Fraction']
) -> Union[VideoFrame, 'ndarray']:
    """
    Apply the video effects to the given
    'frame' on the 't' time moment provided.

    This method should be called before 
    yielding any frame.
    """
    ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)
    ParameterValidator.validate_mandatory_instance_of('effects_stack', effects_stack, VideoEffectsStack)

    if len(effects_stack.effects) == 0:
        return frame

    # TODO: I think this has to preserve the
    # transparency if any of the video frames
    # that are passed to this method (from 
    # the different layers) have an alpha
    # channel or we will lose it in the
    # process... By now I'm just forcing the
    # format that comes. Or maybe not,
    # because when a frame comes and has no
    # transparency, the next frames will be
    # not visible (if coming ordered by
    # priority)... Well... maybe I'm confused
    # right now and this just modifies a 
    # single frame or a frame that has been
    # combined previously, so no worries...

    # We handle frames as 'rgb24' or 'rgba' and
    # then reformat to the expected format (if
    # needed)
    temp_format = (
        'rgba'
        if (
            frame.format.name != 'rgba' and
            'a' in frame.format.name
        ) else
        'rgb24'
        if (
            frame.format.name != 'rgb24' and
            'a' not in frame.format.name
        ) else
        frame.format.name
    )

    # Need to send the frame as a numpy for
    # the effects
    new_frame = effects_stack.apply_effects_at(
        frame = frame.to_ndarray(
            format = temp_format
        ),
        # The 't' here is the internal valid one
        t = t
    )

    """
    When applying the video effects we use
    opengl textures that return, always, an
    alpha channel, so we need to remove it
    if we don't actually need it
    """
    new_frame = (
        RGBAFrameUtils.remove_alpha_channel_from_frame(new_frame)
        if temp_format == 'rgb24' else
        new_frame
    )

    # Rebuild the VideoFrame
    new_frame = VideoFrame.from_ndarray(
        array = new_frame,
        format = temp_format
    )

    new_frame = (
        new_frame.reformat(format = frame.format.name)
        if frame.format.name != temp_format else
        new_frame
    )

    """
    We need 'time_base' and 'pts' values to
    be identified by pyav as valid frames 
    but we don't actually care about the
    values because the Timeline that renders
    will overwrite them
    """
    new_frame.time_base = (
        frame.time_base
        if frame.time_base is not None else
        Fraction(1, 60)
    )

    new_frame.pts = (
        frame.pts
        if frame.pts is not None else
        0
    )

    return new_frame

def apply_audio_effects_to_frame_at(
    effects_stack: AudioEffectsStack,
    frame: Union['AudioFrame', 'ndarray'],
    t: Union[int, float, 'Fraction']
) -> Union['AudioFrame', 'ndarray']:
    """
    Apply the audio effects to the given
    'frame' on the 't' time moment provided.

    This method should be called before 
    yielding any frame.
    """
    # TODO: I think we shouldn't receive a
    # 'ndarray' here, it must be AudioFrame
    ParameterValidator.validate_mandatory_instance_of('frame', frame, [AudioFrame, 'ndarray'])

    # Need the frame as a numpy
    new_frame = (
        frame.to_ndarray()
        if PythonValidator.is_instance_of(frame, AudioFrame) else
        frame
    )
    
    new_frame = effects_stack.apply_effects_at(
        frame = new_frame,
        # The 't' here is the internal valid one
        t = t
    )

    # Rebuild the AudioFrame
    new_frame = AudioFrame.from_ndarray(
        array = new_frame,
        format = frame.format,
        layout = frame.layout
    )

    new_frame.sample_rate = frame.sample_rate

    """
    When applying the video effects we use
    opengl textures that return, always, an
    alpha channel, so we need to remove it
    if we don't actually need it
    """
    new_frame.time_base = (
        frame.time_base
        if frame.time_base is not None else
        Fraction(1, 60)
    )

    new_frame.pts = (
        frame.pts
        if frame.pts is not None else
        0
    )

    return new_frame




from yta_editor.utils.silence import generate_silent_frames
from yta_editor.utils.frame_wrapper import AudioFrameWrapped, VideoFrameWrapped
from yta_editor.utils.frame_generator import VideoFrameGenerator
from yta_video_frame_time.t_fraction import fps_to_time_base
from yta_time_interval import TimeInterval, TIME_INTERVAL_SYSTEM_LIMITS
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union
from abc import ABC, abstractmethod


# TODO: This is reapeated and not used everywhere
Number = Union[int, float, Fraction]
"""
Numeric type we accept as parameter.
"""

class _TrackItem(ABC):
    """
    Abstract class to represent an element that is on
    the track, that can be an empty space, a video, a
    transition or an audio.

    The time moments of this instance will be always a
    multiple of `1/fps`, using the `fps` of the track
    in which the item is located.
    
    This class must be inherited by our own custom item
    classes.
    """

    @property
    def t_start(
        self
    ) -> Fraction:
        """
        The global `t_start` time moment of this gap item in
        which the gap should t_start being played.
        """
        return self._time_interval.t_start

    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The global `t_end` time moment of this gap item in
        which the gap should stop being played.
        """
        return self._time_interval.t_end
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the gap based on its `t_start` and `t_end`
        time moments.

        The formula:
        - `self.t_end - self.t_start`
        """
        return self.t_end - self.t_start

    @property
    def fps(
        self
    ) -> float:
        """
        The fps of the track item, that is the same value as
        the track fps.
        """
        return self._track.fps
    
    @property
    def audio_fps(
        self
    ) -> float:
        """
        The audio fps of the track item, that is the same value
        as the track audio fps.
        """
        return self._track.audio_fps

    @property
    def t_start(
        self
    ) -> Fraction:
        """
        The global `t_start` time moment of this track item, which
        is obtained from the internal time interval.
        """
        return self._time_interval.t_start
    
    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The global `t_end` time moment of this track item, which
        is obtained from the internal time interval.
        """
        return self._time_interval.t_end
    
    @property
    def duration(
        self
    ) -> Fraction:
        """
        The total duration of this track item based on its global
        `t_start` and `t_end` time moments.
        """
        return self._time_interval.duration
    
    @property
    @abstractmethod
    def copy(
        self
    ) -> '_TrackItem':
        """
        A copy of the current instance.
        """
        pass 

    def __init__(
        self,
        track: Union['AudioTrack', 'VideoTrack'],
        t_start: Union[int, float, Fraction],
        duration: Union[int, float, Fraction],
        item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None
    ):
        """
        The `t_start` and `duration` values will be transformed into
        multiples of `1/fps`.
        """
        ParameterValidator.validate_mandatory_positive_number('t_start', t_start, do_include_zero = True)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)

        self._track: Union['AudioTrack', 'VideoTrack'] = track
        """
        The instance of the track this track item belongs to.
        """

        self._time_interval: TimeInterval = TimeInterval(
            t_start = t_start,
            t_end = t_start + duration,
            t_start_limit = None,
            t_end_limit = None,
            # TODO: This 'duration' has to be used when media
            duration_limit = None,
            fps = self.fps
        )
        """
        *For internal use only*

        Internal time interval instance to be able to manage the
        `t_start` and `t_end` time moments properly.
        """

        self.item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
        """
        The previous track item.
        """
        self.item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
        """
        The next track item.
        """
        self.transition_in: Union['TransitionTrackItem'] = None
        """
        The transition that will make the previous item in the
        track be mixed with this one at the begining.
        """
        self.transition_out: Union['TransitionTrackItem'] = None
        """
        The transition that will make the next item in the
        track be mixed with this one at the t_end.
        """

        # We will set it like this to force recalculations if needed
        if item_in is not None:
            self.set_item_in(item_in)
        if item_out is not None:
            self.set_item_out(item_out)
        if transition_in is not None:
            self.set_transition_in(transition_in)
        if transition_out is not None:
            self.set_transition_out(transition_out)

    def set_item_in(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem'],
        do_propagate: bool = True,
    ) -> '_TrackItem':
        """
        Set the provided `item` as the previous one (`.item_in`),
        and this one (self) as the `.item_out` of the one provided
        as parameter, only if the `do_propagate` parameter is True.
        """
        # TODO: Validate as one of those instances and not None
        self.item_in = item

        if (
            item.item_out is not self and
            do_propagate
        ):
            item.set_item_out(
                item = self,
                do_propagate = False
            )
        # TODO: Update the others affected if necessary (?)

        return self
    
    def unset_item_in(
        self,
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Unset the previous item (`.item_in`) from this instance
        but also unset the `.item_out` of the previous one only if
        the `do_propagate` parameter is True.
        """
        if self.item_in is not None:
            item = self.item_in
            self.item_in = None
            # TODO: Setting 'None' maybe should be setting a Gap...

            if (
                item.item_out is self and
                do_propagate
            ):
                item.unset_item_out(do_propagate = False)

        return self
    
    def set_item_out(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem'],
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Set the provided `item` as the next one (`.item_out`),
        and this item (self) as the `.item_in` of the `item`
        provided only if the `do_propagate` parameter is True.
        """
        self.item_out = item

        if (
            item.item_in is not self and
            do_propagate
        ):
            item.set_item_in(
                item = self,
                do_propagate = False
            )
        # TODO: Update the others affected if necessary (?)
        
        return self
    
    def unset_item_out(
        self,
        do_propagate: bool = True
    ) -> '_TrackItem':
        """
        Unset the previous item (`.item_out`) from this instance
        but also unset the `.item_in` of the previous one only if
        the `do_propagate` parameter is True.
        """
        if self.item_out is not None:
            item = self.item_out
            self.item_out = None
            # TODO: Setting 'None' maybe should be setting a Gap...

            if (
                item.item_in is self and
                do_propagate
            ):
                item.unset_item_in(do_propagate = False)

        return self
    
    def set_transition_in(
        self,
        transition: 'TransitionTrackItem'
    ) -> '_TrackItem':
        """
        Set the `transition` provided as the transition to be applied
        to mix the previous item `.item_in` with this one.
        
        TODO: Do the 'decrease' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (decreasing it) according to the duration of the transition
        applied, as the next items will appear earlier.
        """
        if self.item_in is None:
            raise Exception('There is no `.item_in` (previous item) to transition with.')
        
        self.transition_in = transition
        # TODO: We need to update (decrease 'self.duration') the next
        # items starts

        return self
    
    def unset_transition_in(
        self
    ) -> '_TrackItem':
        """
        Unset the transition with the previous item (`.transition_in`).

        TODO: Do the 'reset' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (increasing it) according to the duration of the transition
        applied, as the next items will appear later.
        """
        self.transition_in = None
        # TODO: We need to update (increase 'self.duration') the next
        # items starts

        return self
    
    def set_transition_out(
        self,
        transition: 'TransitionTrackItem'
    ) -> '_TrackItem':
        """
        Set the `transition` provided as the transition to be applied
        to mix the next item `.item_out` with this one.
        
        TODO: Do the 'decrease' here or in the 'in' (?)
        
        This method will update the `t_start` time of the next items 
        (decreasing it) according to the duration of the transition
        applied, as the next items will appear earlier.
        """
        if self.item_out is None:
            raise Exception('There is no `.item_out` (next item) to transition with.')
        
        self.transition_out = transition
        # TODO: The 'decrease' will be done here or in the 'set_transition_in' (?)

        return self
    
    def unset_transition_out(
        self
    ) -> '_TrackItem':
        """
        Unset the transition with the next item (`.transition_out`).

        TODO: Do the 'reset' here or in the 'in' (?)

        This method will update the `t_start` time of the next items 
        (increasing it) according to the duration of the transition
        applied, as the next items will appear later.
        """
        self.transition_out = None
        # TODO: The 'decrease' will be done here or in the 'set_transition_in' (?)

        return self
    
    """
    Functionality related to the time interval below.
    """

    def _shift_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        *For internal use only*

        This method should be called by the `Track` instance that
        holds this item.

        Move the `t_start` and `t_end` value of this item the amount
        of time defined by the `delta` provided parameter, that
        can be positive or negative (and will be truncated to be
        a multiple of `1/fps`).

        The `t_end` value will never go beyond the system's limit,
        so the duration of the item will be enshortened when delta
        is too high.
        """
        if (self.t_end + delta) >= TIME_INTERVAL_SYSTEM_LIMITS[1]:
            self._time_interval.with_start_and_end(
                t_start = self.t_start + delta,
                t_end = TIME_INTERVAL_SYSTEM_LIMITS[1]
            )
        else:
            self._time_interval.shift_by(delta)

        return self
    
    # TODO: Separate the other methods as '_shift_by' and 'shift_by'
    def shift_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Move the `t_start` and `t_end` value of this item to the time
        moment defined by the `t_track` parameter provided. The
        change will not modify the duration of the item.
        """
        return self._shift_by(
            delta = t_track - self.t_start
        )
    
    def shift_start_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_start` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        return self.shift_start_by(
            delta = t_track - self.t_start
        )
    
    def shift_end_by(
        self,
        delta: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value by adding the `delta` provided
        (that can be positive to make it start later or negative
        to make it start earlier).
        """
        self._time_interval.shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Update the `t_end` value to the one provided as `t_track`
        (that can be before or after the current one).
        """
        return self.shift_end_by(
            delta = t_track - self.t_end
        )
    
    def split(
        self,
        t_track: Number
    ) -> '_TrackItem':
        """
        Split the current item at the `t_track` track time
        moment provided and get the 2 new TimeInterval
        instances as a tuple:
        - A: `[t_start, t_track)`
        - B: `[t_track, t_end)`.
        """
        return self._time_interval.split(t_track)
    
    def cut(
        self,
        track_start: Number,
        track_end: Number
    ) -> '_TrackItem':
        """
        Transform this track item into a new one which
        `track_start` and `track_end` are now the ones
        provided as parameters.
        """
        # TODO: What if 't_start' and 't_end' are greater than
        # before? It is not actually cutting...
        self._time_interval.cut(track_start, track_end)

        return self
    
    # TODO: What (?)
    #def propagate_
    
    """
    Functionality related to the time interval above.
    """
    
    def __str__(
        self
    ) -> str:
        """
        Function to stringify the instance and show the in
        and out items next to the item itself.
        """
        item_in: str = (
            f'{self.item_in.__class__.__name__} [{str(float(self.item_in.t_start))}, {str(float(self.item_in.t_end))}]'
            if self.item_in is not None else
            '--'
        )

        item_out: str = (
            f'{self.item_out.__class__.__name__} [{str(float(self.item_out.t_start))}, {str(float(self.item_out.t_end))}]'
            if self.item_out is not None else
            '--'
        )

        return f'  > (In: {item_in}) -   | {self.__class__.__name__} [{str(float(self.t_start))}, {str(float(self.t_end))}] |   - (Out: {item_out})'
    
class _TrackItemWithMedia(_TrackItem):
    """
    Abstract class to represent a track item that includes a
    media file to obtain the frames from. This item is 
    special because its time interval and cannot be shifted
    randomly.

    Here you have some specific cases:
    - If the item is shifted in general, only the item time
    interval will be moved, the media time interval will be
    the same.
    - If the start of the item is shifted, the media's start
    will be also shifted to fit the same duration (this is
    like enshorting or enlarging a video from the left).
    - If the end of the item is shifted, the media's end will
    be also shifted to fit the same duration (this is like
    enshorting or enlarging a video from the right).

    Based on this before, if you need an item to

    """

    @property
    def start_limit(
        self
    ) -> Fraction:
        """
        The global time moment that is the minimum `t` time moment
        this item should be able to handle according to the media
        and the source related to this media.

        Considering a source of 8s, if this item is placed in the
        [9, 11) of the track, it should accept any `t` time moment
        in between those 9 and 11 limits. But, as the media associated
        could be, for example, [1, 3), a `t_track=8` could be accepted
        in special cases (transitions) because it would be transformed
        into `t_media=-1`, which would return `t_source=0` that is
        valid. But not a `t_track=7`, because it would be `t_media=-2`,
        which is `t_source=-1` (invalid).

        This `start_limit` must be used carefully and only for
        special cases (such as transitions).
        """
        # The 'self.media.start_limit' will be always negative or 0
        return self.t_start + self.media.start_limit

    @property
    def end_limit(
        self
    ) -> Fraction:
        """
        The global time moment that is the maximum `t` time moment
        this item should be able to handle according to the media
        and the source related to this media.
        
        Considering a source of 8s, if this item is placed in the
        [9, 11) of the track, it should accept any `t` time moment
        in between those 9 and 11 limits. But, as the media associated
        could be, for example, [2, 4), a `t_track=12` could be accepted
        in special cases (transitions) because it would be transformed
        into `t_media=3`, which would return `t_source=5` that is
        valid. But not a `t_track=16`, because it would be `t_media=7`,
        which is `t_source=9` (invalid).

        This `end_limit` must be used carefully and only for
        special cases (such as transitions).
        """
        return self.t_start + self.media.end_limit

    def __init__(
        self,
        t_start: Number,
        media: Union['AudioMedia', 'VideoMedia'],
        **kwargs
    ):
        # TODO: I think I don't need 'AudioTime' nor 'VideoTimed'
        # but only 'AudioMedia' and 'VideoMedia'
        # ParameterValidator.validate_mandatory_instance_of('media', media, ['AudioMedia', 'VideoMedia'])

        # TODO: I think I don't need 'AudioTime' nor 'VideoTimed'
        # but only 'AudioMedia' and 'VideoMedia'
        self.media: Union['AudioMedia', 'VideoMedia'] = media
        """
        The media associated to this track item.
        """

        super().__init__(
            t_start = t_start,
            duration = media.duration,
            **kwargs
        )
        
        self._time_interval.duration_limit = self.media.max_duration
        
    def _t_track_to_t_media(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Fraction:
        """
        *For internal use only*

        Get the local `t` time moment value for the media
        associated, based on the global `t` time moment
        provided.

        This method will raise an exception if the `t`
        time moment provided is not included in this track
        item time interval and the `do_use_source_limits`
        parameter is not `True`. Send it as `True` if you
        need to force reading the media maybe for transitions.

        The formula:
        - `t_track - self.t_start`
        """
        start_limit = (
            self.start_limit
            if do_use_source_limits else
            self._time_interval.t_start
        )

        end_limit = (
            self.end_limit
            if do_use_source_limits else
            self._time_interval.t_end
        )

        ParameterValidator.validate_mandatory_number_between('t_track', t_track, start_limit, end_limit, True, False)
        
        """
        Imagine that we have this item in [4, 8] and we
        receive t=6. The local_t=6-4=2. The media we have
        available has 4s of duration (so [0, 4]), but it
        is actually cropped to [1, 3], so the moment we
        need to read is the t=2 of the cropped version,
        so: t_media=1+2=3. But is the media who will
        calculate that internal t, we only need to ask him
        for the t=2 and he will make the conversion (that
        is currently being done by a decorator).

        That means that we could actually ask for the t=-1
        to the media, that would be turned into the t=0 of
        the source which is actually available, so we have
        to be very careful with this...

        Source [0, 8)
        Media [0, 2) but of the source [2, 4)
        TrackItem [11, 13)
        -> Ask t_t=12 => t_media=1 => t_source = 3
            Is valid and included in the normal limits
        -> Ask t_t=10 => t_media=-1 => t_source=1
            Is valid only if forcing to use the source
            limits because t_source=1, but t_t=10 is out
            of the [11, 13) time interval
        """

        return t_track - self.t_start
    
    # The super() methods will shift only the item
    """
    These methods below will shift the media only.
    """
    def shift_media_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` and `t_end` time moments by
        the `delta` amount of time provided.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_media_by(1)`.
        The item will be still [1, 4), but the media will be
        pointing to the [3, 6) time interval of the source.
        """
        self.media.shift_by(delta)
    
        return self
    
    def shift_media_start_and_item_end_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` time moment by the `delta`
        amount of time provided, but the `t_end` of the media is
        not affected, and updates the item `t_end` time moment by
        that same `delta` amount of time, affecting not to the
        item's `t_start` time moment.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_media_start_and_item_end_by(1)`. The item will be
        [1, 3), but the media will be pointing to the [4, 6) time
        interval of the source.
        """
        self.media.shift_start_by(delta)

        self._time_interval.shift_end_by(-delta)

        return self

    def shift_media_to(
        self,
        t_media: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the media's `t_start` to the `t_media` time moment
        provided, and the `t_end` the same amount of time.

        I'll give you one example below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_media_to(3)`.
        The item will be still [1, 4), but the media will be
        pointing to the [3, 6) time interval of the source.
        """
        delta = t_media - self.media.t_start

        return self.shift_media_by(delta)

    """
    These methods below will shift both the item and the media.
    """
    def shift_item_and_media_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        TODO: Shifting item and media at the same time? I don't
        think so... (?)

        Update the item `t_start` time moment by the `delta` value
        provided but also the media's `t_start` by the same `delta`
        value, and the `t_end` time moment of both will be also
        modified with the same amount of time. This value can be
        positive (will make it start later) or negative (will make
        it start earlier).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_by(1)`. The item will be now [2, 5),
        and the media will be pointing to the [3, 6) time interval
        of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do
        `.shift_item_and_media_by(-1)`. The item will be now [0, 3),
        and the media will be pointing to the [1, 4) time interval
        of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_by(-3)`. The item can be [0, 6)
        because the duration of the media is 8s, but the media would
        be pointing to the source's [-1, 2) time interval, which is
        not possible because is out of the limits (minimum `t_start
        value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_by(delta)

        self._shift_by(delta)

        return self

    def shift_item_and_media_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        TODO: Shifting item and media at the same time? I don't
        think so... (?)

        Update the item `t_start` time moment to the `t_track` time
        moment provided but also the media's `t_start` time 
        variation, and the `t_end` time moment of both will be also
        modified with the same amount of time. This value can be
        greater (will make it start later) or lower (will make
        it start earlier) than the current track `t_start` time
        moment.

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_to(2)`. The item will be now [2, 5),
        and the media will be pointing to the [3, 6) time interval
        of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do
        `.shift_item_and_media_to(0)`. The item will be now [0, 3),
        and the media will be pointing to the [1, 4) time interval
        of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do
        `.shift_item_and_media_to(0)`. The item can be [0, 6)
        because the duration of the media is 8s, but the media would
        be pointing to the source's [-1, 2) time interval, which is
        not possible because is out of the limits (minimum `t_start
        value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_start
        
        return self.shift_item_and_media_by(delta)

    """
    These methods below will shift the start and the end
    of the item and the media at the same time.
    """
    def shift_start_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_start` time moment by the `delta` value
        provided but also the media's `t_start` by the same `delta`
        value. This value can be positive (will make it last less
        time) or negative (will make it longer).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_by(1)`.
        The item will be now [2, 4), and the media will be pointing
        to the [3, 5) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_start_by(-1)`.
        The item will be now [0, 4), and the media will be pointing
        to the [1, 5) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_by(-3)`.
        The item can be [0, 6) because the duration of the media is
        8s, but the media would be pointing to the source's [-1, 2)
        time interval, which is not possible because is out of the
        limits (minimum `t_start` value is 0)..
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_start_by(delta)

        super().shift_start_by(delta)

        return self
    
    def shift_start_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_start` time moment to the `t_track` time
        moment value provided, but also the media the same amount
        of time than the item (only if possible). Be careful with
        the limits of the media's time interval.

        I'll give you some examples below with a media with 8s of
        duration:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(3)`.
        The item will be now [3, 4), and the media will be pointing
        to the [4, 5) of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(0)`.
        The item will be now [0, 4), and the media will be pointing
        to the [1, 5) of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_start_to(0)`.
        The item can be [0, 6) because the duration of the media is
        8s, but the media would be pointing to the source's [-1, 2)
        time interval, which is not possible because is out of the
        limits (minimum `t_start` value is 0).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_start

        # This will force the 2 shifts according to their limits
        return self.shift_start_by(delta)
    
    def shift_end_by(
        self,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_end` time moment by the `delta` value
        provided but also the media's `t_end` by the same `delta`
        value. This value can be positive (will make it last more
        time) or negative (will make it shorter).

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_by(1)`.
        The item will be now [1, 5), and the media will be pointing
        to the [2, 6) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_end_by(-1)`.
        The item will be now [1, 3), and the media will be pointing
        to the [2, 4) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_by(4)`.
        The item can be [3, 10) because the duration of the media is
        8s, but the media would be pointing to the source's [2, 9)
        time interval, which is not possible because is out of the
        limits (maximum `t_end` value is 8).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        # TODO: Validate that the new `start` is valid according
        # to the item time interval limits
        # TODO: Validate that the new `media_start` is valid
        # according to the media interval limit
        # TODO: Is this above validated when 'shifting' (?)
        self.media.shift_end_by(delta)

        super().shift_end_by(delta)

        return self
    
    def shift_end_to(
        self,
        t_track: Union[int, float, Fraction]
    ) -> '_TrackItemWithMedia':
        """
        Update the item `t_end` time moment to the `t_track` time
        moment value provided, but also the media the same amount
        of time than the item (only if possible). Be careful with
        the limits of the media's time interval.

        I'll give you some examples below:
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_to(5)`.
        The item will be now [1, 5), and the media will be pointing
        to the [2, 6) time interval of the source.
        - We have the item at [1, 4), and the media is pointing to
        the [2, 5) of the media, and we do `.shift_end_to(3)`.
        The item will be now [1, 3), and the media will be pointing
        to the [2, 4) time interval of the source.
        - We have the item at [3, 6), and the media is pointing to
        the [2, 5) of the source, and we do `.shift_end_to(10)`.
        The item can be [3, 10) because the duration of the media is
        8s, but the media would be pointing to the source's [2, 9)
        time interval, which is not possible because is out of the
        limits (maximum `t_end` value is 8).
        
        This is limited, at first, by the item time interval limit
        but also by the media time interval limit.
        """
        delta = t_track - self.t_end

        # This will force the 2 shifts according to their limits
        return self.shift_end_by(delta)
    
    # TODO: We need these methods:
    # - shift_only_item: moves the item time interval but not media
    # - shift: moves the item time interval but also the media
    # - shift_only_media: moves the media time interval only
    # Any other method (shift_start, shift_end) will shift both
    # because it is necessary to do it

class _TrackItemWithAudio(_TrackItem):
    """
    Abstract class to implement a track item that is capable
    of providing audio frames.
    """

    @property
    def _silent_frames(
        self
    ) -> list[AudioFrameWrapped]:
        """
        *For internal use only*

        The silent frames we need to fulfill the
        space when we didn't find the audio frames
        or the part of the video is just silent.

        This property has been created to avoid
        creating these silent frames again and again,
        as they will be always the same.
        """
        if not hasattr(self, '__silent_frames'):
            self.__silent_frames = generate_silent_frames(
                video_fps = self._track.fps,
                audio_fps = self._track.audio_fps,
                audio_samples_per_frame = self._track.audio_samples_per_frame,
                # TODO: Where do this 2 formats come from (?)
                layout = self._track.audio_layout,
                format = self._track.audio_format
            )

        return self.__silent_frames
    
    # def __init__(
    #     # track: 'AudioTrack',
    #     **kwargs
    # ):
    #     # ParameterValidator.validate_mandatory_instance_of('track', track, 'AudioTrack')

    #     super().__init__(
    #         # track = track,
    #         **kwargs
    #     )

    def get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have the 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Iterate over all the audio frames that
        exist at the time moment 't' provided.
        """
        # frames = []
        # # TODO: This should be different with the new gap item
        # if not self.is_empty_part:
        #     # TODO: What do we do in this case (?)
        #     frames = list(self._get_audio_frames_at(t_track))

        #     if len(frames) == 0:
        #         print(f'   [ERROR] Audio frame {str(float(t_track))} was not obtained')
        #     else:
        #         frames = [
        #             AudioFrameWrapped(
        #                 frame = frame,
        #                 is_from_empty_part = False
        #             )
        #             for frame in frames
        #         ]
        # TODO: Remove this on top if below is working

        frames = list(self._get_audio_frames_at(
            t_track = t_track,
            video_fps = video_fps,
            do_use_source_limits = do_use_source_limits
        ))

        if len(frames) == 0:
            print(f'   [ERROR] Audio frame {str(float(t_track))} was not obtained')
        else:
            # TODO: We should not receive AudioFrame and 
            # AudioFrameWrapped here, only one. Fix it (!)
            from yta_validation import PythonValidator
            frames = [
                (
                    AudioFrameWrapped(
                        frame = frame,
                        is_from_empty_part = False
                    )
                )
                if PythonValidator.is_instance_of(frame, 'AudioFrame') else
                frame
                for frame in frames
            ]

        # This could be because is empty part or
        # because we couldn't obtain the frames
        frames = (
            self._silent_frames
            if len(frames) == 0 else
            frames
        )

        for frame in frames:
            yield frame

    @abstractmethod
    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames of the provided `t` global time
        moment according to the internal configuration of 
        the track item.

        TODO: This method must be overwritten by the
        specific class implementations.
        """
        pass

class TrackItemWithAudioMedia(_TrackItemWithAudio, _TrackItemWithMedia):
    """
    A track item that is capable of providing audio
    frames by obtaining them from the media it contains.
    """

    def __init__(
        self,
        track: 'AudioTrack',
        t_start: Union[int, float, Fraction],
        media: 'AudioMedia',
        item_in: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None
    ):
        super().__init__(
            track = track,
            t_start = t_start,
            media = media,
            item_in = item_in,
            item_out = item_out,
            transition_in = transition_in,
            transition_out = transition_out
        )

    def __init__(
        self,
        media: 'AudioTimed',
        **kwargs
    ):
        ParameterValidator.validate_mandatory_instance_of('media', media, 'AudioTimed')

        super().__init__(
            media = media,
            **kwargs
        )

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames for the provided `t` global
        time moment, that is transformed into the media
        local time moment, from the audio media this
        instance contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )
        
        return self.media.get_audio_frames_at(
            t_media = t_media,
            video_fps = self.fps,
            do_use_source_limits = do_use_source_limits
        )
    
class _TrackItemWithVideo(_TrackItem):
    """
    Abstract class to implement a track item that is capable
    of providing audio frames.
    """

    @property
    def _frame_not_found(
        self
    ):
        """
        *For internal use only*

        The frame we want to send to the editor when
        we were not able to find the frame we were
        looking for (due to an error or that the file
        was unreadable in that specific moment).

        This property has been created to avoid
        creating these empty frames again and again,
        as they will be always the same.

        TODO: This should not happen, it is an
        internal way to point an error, and of
        course should be removed in a future 
        version when we don't suffer it.
        """
        if not hasattr(self, '__frame_not_found'):
            self.__frame_not_found = VideoFrameGenerator.background.full_red(
                size = self._track.size,
                time_base = fps_to_time_base(self._track.fps)
            )

        return self.__frame_not_found

    # def __init__(
    #     self,
    #     # track: 'VideoTrack',
    #     **kwargs
    # ):
    #     # ParameterValidator.validate_mandatory_instance_of('track', track, 'VideoTrack')

    #     super().__init__(
    #         # track = track,
    #         **kwargs
    #     )

    def __init__(
        self,
        is_empty_part: bool = False,
        **kwargs
    ):
        super().__init__(
            **kwargs
        )

        self._is_empty_part: bool = is_empty_part
        """
        *For internal use only*

        Internal flag to indicate if this track item is an
        empty part or not.

        TODO: This is a remaining part of the previous way
        to handle the empty parts, so maybe we can remove
        it soon.
        """

    # TODO: This method was previously 'get_video_frame_at' so
    # refactor its use
    def get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> 'VideoFrameWrapped':
        """
        Get the frame that must be displayed at 
        the given 't_track' global time moment.
        """
        frame: Union['VideoFrame', None] = self._get_video_frame_at(t_track, do_use_source_limits)
        # TODO: This can be None, why? I don't know...
        # TODO: Remove this below, its to be copied...
        # frame = self.media.get_video_frame_at(t)

        if frame is None:
            print(f'   [ERROR] Frame {str(float(t_track))} was not obtained')

        frame = (
            # I'm using a red full frame to be able to detect
            # fast the frames that were not available, but
            # I need to find the error and find a real solution
            # TODO: This shouldn't happen, its an error
            self._frame_not_found
            if frame is None else
            frame
        )

        # TODO: What about the 'format' (?)
        # TODO: Maybe I shouldn't set the 'time_base'
        # here and do it just in the Timeline 'render'
        #return get_black_background_video_frame(self._track.size)
        # TODO: This 'time_base' maybe has to be related
        # to a Timeline general 'time_base' and not the fps

        # TODO: We need to remove the 'is_from_empty_part' maybe (?)
        frame = VideoFrameWrapped(
            frame = frame,
            is_from_empty_part = self._is_empty_part
        )

        # TODO: This should not happen because of
        # the way we handle the videos here but the
        # video could send us a None frame here, so
        # do we raise exception (?)
        if frame._frame is None:
            #frame = get_black_background_video_frame(self._track.size)
            # TODO: By now I'm raising exception to check if
            # this happens or not because I think it would
            # be malfunctioning
            raise Exception(f'Video is returning None video frame at t={str(t_track)}.')
        
        """
        The 'track' (because of the 'timeline' it belongs
        to) has a size, so we must respect, so we need to
        use a strategy to make it fit the desired size.
        """
        # TODO: Maybe this sould be done in the 'timeline'
        # that combines the frames and not here...
        return frame
    
    @abstractmethod
    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrame', None]:
        """
        Get the video frame for the `t_track` time moment
        provided according to the internal configuration of
        this instance.
        """
        pass

class TrackItemWithVideoMedia(_TrackItemWithVideo, _TrackItemWithAudio, _TrackItemWithMedia):
    """
    A track item that is capable of providing video and
    audio frames by obtaining them from the video media
    it contains.
    """

    @property
    def copy(
        self
    ) -> 'TrackItemWithVideoMedia':
        """
        A copy of the current track item with media
        instance.
        """
        return TrackItemWithVideoMedia(
            track = self._track,
            t_start = self.t_start,
            media = self.media,
            item_in = self.item_in,
            item_out = self.item_out,
            transition_in = self.transition_in,
            transition_out = self.transition_out
        )

    def __init__(
        self,
        track: 'VideoTrack',
        t_start: Union[int, float, Fraction],
        media: 'VideoMedia',
        item_in: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        transition_in: Union['TransitionTrackItem', None] = None,
        transition_out: Union['TransitionTrackItem', None] = None,
    ):
        super().__init__(
            track = track,
            t_start = t_start,
            media = media,
            item_in = item_in,
            item_out = item_out,
            transition_in = transition_in,
            transition_out = transition_out,
        )

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        do_use_source_limits: bool = False
    ):
        """
        Get the audio frames for the provided `t_track`
        global time moment from the audio media this instance
        contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )

        return self.media.get_audio_frames_at(
            t_media = t_media,
            video_fps = self.fps,
            do_use_source_limits = do_use_source_limits
        )

    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction],
        do_use_source_limits: bool = False
    ) -> Union['VideoFrame', None]:
        """
        Get the video frame for the provided `t_track` time
        moment from the video media this instance contains.

        The `do_use_source_limits` method should be used
        when we want to access to a `t_track` that is out
        of the time interval of this track item but it is
        actually accepted by the internal source time
        interval limits.
        """
        t_media = self._t_track_to_t_media(
            t_track = t_track,
            do_use_source_limits = do_use_source_limits
        )

        return self.media.get_video_frame_at(
            t_media = t_media,
            do_use_source_limits = do_use_source_limits
        )
    

"""
Note for the developers:
- We have 3 different types of items:
    *Item part: An item (there are different options) 
    that is occupying that part of the track, being a
    media item, a transition item...
    *Gap part in between: An empty part that is before
    the first media or in between 2 different medias.
    *Gap part at the t_end: An empty part that is at the
    t_end of the track, after the last media, making the
    track virtually infinite.

Check the 'Track' class to see the interaction and
creation.
"""


# from yta_editor.tracks.items.abstract import TrackItemWithAudioMedia, _TrackItemWithVideo
from yta_editor.tracks.items.video import _VideoTrackItem
from yta_editor.tracks.items.abstract import _TrackItemWithVideo, _TrackItemWithAudio
from yta_constants.enum import YTAEnum as Enum
from yta_validation import PythonValidator
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


# TODO: Move it
class TransitionMode(Enum):
    """
    The modes the TransitionTrackItem accept as mode to
    implement it, that is the way you handle the frames
    of the previous and the next clip.
    """

    TRIM = 'trim'
    """
    The transition will use the real frames of both the
    previous and the next clip, which will make both 
    clips last less time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA_trimmed] [Transition] [ClipB_trimmed]`
    """
    FREEZE_TAIL = 'freeze_tail'
    """
    The transition will use the last frame of the 
    previous clip but frozen, which means that it will
    be used for the whole transition duration. The next
    clip 'start' time moment will be increased because
    the previous clip is not trimmed and the transition
    will last an additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """
    FREEZE_HEAD = 'freeze_head'
    """
    The transition will use the first frame of the next
    clip but frozen, which means that it will be used
    for the whole transition duration. The next clip
    'start' time moment will be increased because the
    previous clip is not trimmed and the transition will
    last an additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """
    FREEZE_BOTH = 'freeze_both'
    """
    The transition will use the last frame of the
    previous clip and the first frame of the next one,
    but frozen, which means that they will be used for
    the whole transition duration. The next clip 'start'
    time moment will be increased because the previous
    clip is not trimmed and the transition will last an
    additional amount of time.

    It means that from:
    - `[ClipA] [Gap] [ClipB]`
    We will move to:
    - `[ClipA] [Transition] [ClipB_delayed]`
    """

class TransitionTrackItem(_TrackItemWithVideo, _TrackItemWithAudio):
    """
    A track item that is built by joining 2 different video
    items that are consecutive.
    """

    @property
    def copy(
        self
    ) -> '_TrackItem':
        """
        A copy of the current instance.
        """
        return TransitionTrackItem(
            track = self._track,
            t_start = self.t_start,
            duration = self.duration,
            type = self.type,
            mode = self.mode,
            item_in = self.item_in,
            item_out = self.item_out
        )

    def __init__(
        self,
        track: 'VideoTrack',
        t_start: Union[int, float, Fraction],
        # TODO: We need to be able to handle options
        # different than the 50% of each clip
        duration: Union[int, float, Fraction],
        # TODO: How do we handle the type? Maybe a full config instead (?)
        type: str,
        mode: TransitionMode = TransitionMode.TRIM,
        # TODO: This is a special one so this should be different
        item_in: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None,
        item_out: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem', None] = None
    ):
        mode = TransitionMode.to_enum(mode)

        super().__init__(
            track = track,
            t_start = t_start,
            duration = duration,
            item_in = item_in,
            item_out = item_out,
            # TODO: I think I don't need these transition fields
            # in this kind of item
            transition_in = None,
            transition_out = None
        )

        self.mode: TransitionMode = mode
        """
        The mode in which the transition must be implemented.
        """
        self._transition_calculator: _TransitionCalculator = _TransitionCalculator(
            duration = self.duration
            #  TODO: The 'rate_function' has to come in the configuration
        )
        """
        *For internal use only*

        A transition calculator to simplify the way we handle
        the calculations to obtain the frames we need.
        """

    def _get_audio_frames_at(
        self,
        t_track: Union[int, float, Fraction],
        # TODO: Ignore this, we have 'self.fps' here
        video_fps: Union[int, float, Fraction] = None,
        # TODO: Ignore this, we don't need it...
        do_use_source_limits: bool = True
    ) -> Union['AudioFrame', None]:
        """
        Get the audio frames of the provided `t_track`
        global time moment according to the internal
        configuration of the track item.

        TODO: This method must be overwritten by the
        specific class implementations.
        """
        item_in_audio_frames = self.item_in.get_audio_frames_at(
            t_track = t_track,
            # TODO: We should not receive the param
            do_use_source_limits = True
        )
        item_out_audio_frames = self.item_out.get_audio_frames_at(
            # t_track = t_track + self.duration,
            t_track = t_track,
            # TODO: We should not receive the param
            do_use_source_limits = True
        )

        # Obtain the local 't_track' of the transition (for the progress)
        transition_t = t_track - (self.t_end - self.duration)

        # TODO: Implement this as a combination of the audio
        # coming from from clips at the same time, but by now
        # I'm just returning something
        return item_in_audio_frames

    def _get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction]
    ) -> Union['VideoFrame', None]:
        """
        Get the frame of the transition for the global
        `t_track` time moment provided.
        """
        # TODO: The 't' must be the global value of the timeline
        # Obtain the frames from each video
        # TODO: I actually need to obtain a frame that will be
        # out of the current 'item_in' valid time interval, as
        # it has been trimmed because of this transition, and 
        # the part we are using is just after (for the
        # 'item_in', before for the 'item_out') the last part
        # we are showing in the video
        # TODO: Calculate the 't' based on the 'item_in' t_end
        # TODO: We need to read the 'item_in.t_end' + 'local_t'
        # for the transition

        """
        Imagine this context:
        - item_in: [0, 2)   Transition: [2, 3)   item_out: [3, 4)
        The global t=2.5 is actually the 0.5 inside the 
        transition, that should read the 'item_in.t_end+0.5' (out
        of the time interval that is being shown) of the item_in.

        Now imagine that the item_in is [0, 2) but the media 
        inside is actually [0, 8) and it is reading from [3, 5).
        That means that the transition for the global t=2.5, 
        is a local_t=0.5, should be reading the t=2.5 (out of
        time interval), that is actually the t_media=5.5.
        """

        """
        As the media has been enshortened due to the transition,
        we need to use `t_track` values that are out of the 
        current item media limits, thats why we force the values
        to be allowed with the `do_use_source_limits` parameter.
        """
        frame_item_in = self.item_in.get_video_frame_at(
            t_track = t_track,
            do_use_source_limits = True
        )

        frame_item_out = self.item_out.get_video_frame_at(
            #t_track = t_track + self.duration,
            t_track = t_track,
            do_use_source_limits = True
        )

        # We don't want 'VideoFrameWrapped' to process
        frame_item_in = frame_item_in.as_rgba_numpy
        frame_item_out = frame_item_out.as_rgba_numpy

        # Obtain the local 't' of the transition (for the progress)
        local_t = t_track - self.t_start

        """
        *Representacion con t(global) de clips*
        Clip A: [0, 3)   Clip B: [3, 5)
        - Añadimos transition de 1s
        Clip A: [0, 2)   Transition: [2, 3)   Clip B: [3, 4)
        - La 'Transition' tiene [2, 3) de A y [0, 1) de B
        - El t(global)=2.5 ahora cae en la transición, que se
          convierte a t(local)=2.5-(3.0-1.0)=0.5, y dado que
          la duration es 1.0; 0.5 es el t_progress=0.5
        """

        transition_calculator = _TransitionCalculator(
            duration = self.duration
            #  TODO: The 'rate_function' has to come in the configuration
        )

        t_progress = transition_calculator.get_progress_at(local_t)

        frame = transition_calculator._process_frame(
            frame_a = frame_item_in,
            frame_b = frame_item_out,
            t_progress = t_progress
        )

        # TODO: Improve this
        from yta_editor.utils.frame_wrapper import VideoFrameWrapped

        return VideoFrameWrapped(frame)
    
    # TODO: What if I have another abstract class to inherit
    # from that is different having not the transitions 
    # instead of blocking these methods here? Maybe a 
    # TrackItem, TrackItemWithTransitions (?)
    def set_transition_in(self, transition):
        raise Exception('This class can not have transitions')
    
    def set_transition_out(self, transition):
        raise Exception('This class can not have transitions')
    
    def unset_transition_in(self):
        raise Exception('This class can not have transitions')
    
    def unset_transition_out(self):
        raise Exception('This class can not have transitions')  






from yta_math.rate_functions.rate_function import RateFunction
from yta_editor_nodes.processor.video.transitions import CrossfadeTransitionProcessor
from av.video.frame import VideoFrame

import numpy as np

# TODO: This class below could be repeated and I just
# created because I needed to (at least by now)
# TODO: If I keep this calculator method, apply it in
# the '_TransitionMedia' class if I also keep it
class _TransitionCalculator:
    """
    *For internal use only*

    A class to simplify the way we make the calculations
    related to a transition. This class must be used within
    a transition item that needs these kind of calculations.
    """

    def __init__(
        self,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        self.duration: float = duration
        """
        The duration of the transition.
        """
        self.rate_function: RateFunction = RateFunction.to_enum(rate_function)
        """
        The rate function to be applied in the transition
        progress to handle its speed.
        """
        # TODO: This depends on each node implementation and the
        # arguments are different. By now I'm forcing this
        self._transition_node: CrossfadeTransitionProcessor = CrossfadeTransitionProcessor(
            do_use_gpu = True,
            output_size = (1920, 1080)
        )
        """
        The node that will process the transition, that must
        be implemented by the child class.
        """

    def get_progress_at(
        self,
        t: float
    ) -> float:
        """
        Get the transition progress value, that will
        be in the `[0.0, 1.0]` range, according to the
        `t` time moment provided and the `rate_function`
        that must be applied in this transition.

        This method should be called only when the
        transition has to be applied, so the `t` is
        a time moment in which the two frames are 
        being played together.

        The `t` value here must be a value in the range
        [0, self.duration] and it must be calculated
        before calling this method.
        """

        """
        By now, the transition we are applying is
        as simple as being in the middle of the 2
        clips provided and lasting the 'duration'
        provided, being executed with a linear t.
        """
        # TODO: We force it, by now, to be calculated
        # outside as in this [0, self.duration] range
        ParameterValidator.validate_mandatory_number_between('t', t, 0, self.duration)

        return (
            0.0
            if t < 0 else
            1.0
            if t > self.duration else
            self.rate_function.get_n_value(
                n = max(
                    0.0,
                    min(
                        1.0,
                        t / self.duration
                    )
                )
            )
        )
    
    # TODO: This method can be overwritten to change
    # its behaviour if the specific transition needs
    # it
    def _process_frame(
        self,
        frame_a: Union['moderngl.Texture', 'np.ndarray'],
        frame_b: Union['moderngl.Texture', 'np.ndarray'],
        t_progress: float,
    ):
        # TODO: Maybe this can be placed in the general
        # class if it doesn't change

        # TODO: This can be a texture
        processed_frame = self._get_process_frame_array(
            frame_a = frame_a,
            frame_b = frame_b,
            t_progress = t_progress
        # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
        )

        if PythonValidator.is_numpy_array(processed_frame):
            processed_frame = processed_frame.astype(np.uint8)

            # frame = VideoFrame.from_ndarray(
            #     # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
            #     array = processed_frame,
            #     format = 'rgb24'
            # )
        elif PythonValidator.is_instance_of(processed_frame, 'Texture'):
            # buffer = ctx.buffer(reserve=texture.width * texture.height * 3)
            # texture.read_into(buffer, components=3, dtype='u1')

            # import numpy as np
            # rgb = np.frombuffer(buffer.read(), dtype=np.uint8).reshape((h, w, 3))
            # TODO: Apparently, there is another way to use a reusable
            # buffer to transform a texture into a numpy array, so 
            # please refactor when possible (check the code above)
            # from moderngl import Texture
            # processed_frame: Texture
            #data = processed_frame.read(components = 3, dtype = 'u1')

            # OpenGL uses textures flipped, so maybe (?)
            data = processed_frame.read()
            w, h = processed_frame.size

            # data = np.flip(data, axis = 0)

            # TODO: I'm removing the alpha channel here...
            data = np.frombuffer(data, dtype = np.uint8).reshape((h, w, 4))
            data = np.flip(data, axis = 0)
            processed_frame  = data[:, :, :3]
        else:
            # TODO: Wtf? This should not happen
            raise Exception('Wowowowooo, no texture, no np.ndarray... wtf?')
        
        frame = VideoFrame.from_ndarray(
            # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
            array = processed_frame,
            format = 'rgb24'
        )

        # The 'pts' and that is not important here but...
        frame.pts = None
        frame.time_base = Fraction(1, 60)

        return frame
    
    # TODO: Overwrite this method to make it custom (?)
    def _get_process_frame_array(
        self,
        frame_a: Union['moderngl.Texture', 'np.ndarray'],
        frame_b: Union['moderngl.Texture', 'np.ndarray'],
        t_progress: float,
    ) -> np.ndarray:
        """
        Get the numpy array that will be used to build the
        pyav VideoFrame that will be returned.
        """
        return self._transition_node.process(
            first_input = frame_a,
            second_input = frame_b,
            # TODO: I'm forcing this due to a problem with the uniforms
            progress = float(t_progress)
        )



# from yta_editor.tracks.media.video import VideoTimed
# from yta_editor.media.video import _VideoMedia
# from yta_video_frame_time.t_fraction import get_ts, fps_to_time_base, T
# from yta_math.rate_functions.rate_function import RateFunction
# from yta_video_pyav.writer import VideoWriter
# from yta_video_pyav.settings import Settings
# from yta_validation.parameter import ParameterValidator
# from av.video.frame import VideoFrame
# from quicktions import Fraction
# from typing import Union

# import numpy as np


# class _TransitionMedia:
#     """
#     *For internal use only*

#     This class must be inherited by the transitions
#     that have the specific implementation.

#     Class to wrap a transition clip, that is a
#     clip in which 2 clips are joined with some logic.

#     TODO: This class, by now, is only handling clips
#     that are played together during the 'transition'
#     period of time. No frozen frames, no diferent
#     time moments. When the transition is executed
#     both videos are being played at the same time.
#     """

#     @property
#     def _t_start(
#         self
#     ) -> Fraction:
#         """
#         *For internal use only*

#         The moment in which the transition clip starts
#         being played, that will be in some moment of
#         the first clip, and it should be also the 
#         moment in which the second clip starts being
#         played (with the corresponding transition
#         blending).
#         """
#         # TODO: This value is only valid for the transition
#         # applied in the middle with 50%-50% strategy
#         return self._clip_a.duration - self.duration
    
#     @property
#     def _t_end(
#         self
#     ) -> Fraction:
#         """
#         *For internal use only*

#         The moment in which the whole clip built with
#         the transition in the middle must stop being
#         played, so the second clip has finished.
#         """
#         return self._clip_b.end

#     def __init__(
#         self,
#         # TODO: This '_MediaTimed' has to be renamed 
#         # because here it is not on a track but the class
#         # we need is this one
#         # TODO: Maybe it has to be just a Media, not OnTrack
#         clip_a: _VideoMedia,
#         clip_b: _VideoMedia,
#         duration: float,
#         rate_function: RateFunction = RateFunction.LINEAR,
#     ):
#         """
#         The clips must come ordered, so the `clip_a` must be
#         playing before the `clip_b`.
#         """
#         print(type(clip_a))
#         # TODO: Here we are receiving VideoTimed
#         ParameterValidator.validate_subclass_of('clip_a', clip_a, _VideoMedia)
#         ParameterValidator.validate_subclass_of('clip_b', clip_b, _VideoMedia)

#         if clip_b.start > clip_a.start:
#             raise Exception('The `clip_a` must start before the `clip_b`.')

#         if (
#             clip_a.duration < duration or
#             clip_b.duration < duration
#         ):
#             raise Exception('The duration of one of the clips is smaller than the transition duration requested.')

#         # TODO: Apply Union[AlphaBlendTransition, ...]
#         self.duration: float = duration
#         """
#         The duration of the transition.
#         """
#         self.rate_function: RateFunction = RateFunction.to_enum(rate_function)
#         """
#         The rate function to be applied in the transition
#         progress to handle its speed.
#         """

#         self._clip_a: VideoTimed = VideoTimed(
#             media = clip_a,
#             start = 0
#         )
#         """
#         The first clip to join in with the transition.
#         """
#         self._clip_b: VideoTimed = VideoTimed(
#             media = clip_b,
#             start = self._t_start
#         )
#         """
#         The second clip to join in with the transition.
#         """

#         self._transition_node: 'TransitionNode' = None
#         """
#         The node that will process the transition, that must
#         be implemented by the child class.
#         """

#         # TODO: Set '_transition_node' in children and remove
#         # this below
#         # TODO: The child class must implement the
#         # 'self.opengl_node' variable
#         # self.opengl_node: '_OpenglNodeBase' = CrossfadeTransitionNode(
#         #     opengl_context = OpenGLContext().context,
#         #     # TODO: Do not hardcode, please
#         #     size = (1920, 1080),
#         # )
    
#     def get_transition_progress_at(
#         self,
#         t: float
#     ) -> float:
#         """
#         Get the transition progress value, that will
#         be in the [0.0, 1.0] range, according to the
#         `t` time moment provided and the `rate_function`
#         that must be applied in this transition.

#         This method should be called only when the
#         transition has to be applied, so the `t` is
#         a time moment in which the two frames are 
#         being played together.
#         """

#         """
#         By now, the transition we are applying is
#         as simple as being in the middle of the 2
#         clips provided and lasting the 'duration'
#         provided, being executed with a linear t.
#         """
#         # The transition can start in a time moment
#         # different than 0 so we recalculate it
#         t -= self._t_start

#         return (
#             0.0
#             if t < 0 else
#             1.0
#             if t > self.duration else
#             self.rate_function.get_n_value(
#                 n = max(
#                     0.0,
#                     min(
#                         1.0,
#                         t / self.duration
#                     )
#                 )
#             )
#         )
    
#     def get_video_frame_at(
#         self,
#         t: Union[int, float, Fraction],
#         do_apply_filters: bool = False
#     ) -> VideoFrame:
#         """
#         The `t` time moment provided must be a global
#         time value pointing to a moment in the general
#         timeline. That value will be transformed to the
#         internal `t` value to get the frame.
#         """
#         # Test with OpenGL
#         frame_a = self._clip_a.get_video_frame_at(t)
#         frame_b = self._clip_b.get_video_frame_at(t)

#         if (
#             frame_a is None and
#             frame_b is None
#         ):
#             print('Both None, wtf')
#             #return None

#         """
#         There is a moment in which we are not in
#         the transition section but on the clips
#         so we just need to return the clip frames.
#         """
#         if frame_b is None:
#             # Not transition time, just one clip
#             out_frame = frame_a
#             out_frame.pts = None
#             out_frame.time_base = Fraction(1, 60)

#             return out_frame
        
#         if frame_a is None:
#             # Not transition time, just one clip
#             out_frame = frame_b
#             out_frame.pts = None
#             out_frame.time_base = Fraction(1, 60)

#             return out_frame
        
#         # Render loop
#         # TODO: Why as 'rgb24' and 'np.float32' and no 'np.uint8' (?)
#         # TODO: We should try to read with GPU maybe
#         # TODO: What if it comes from a numpy and not a file as VideoFrame (?)
#         frame_a = frame_a.to_ndarray(format = 'rgb24').astype(np.float32)
#         frame_b = frame_b.to_ndarray(format = 'rgb24').astype(np.float32)

#         # TODO: Do we need the alpha (not rgb24?) maybe
#         # to mix with something else
#         return self._process_frame(
#             frame_a = frame_a,
#             frame_b = frame_b,
#             t_progress = self.get_transition_progress_at(t)
#         )
    
#     # TODO: This method can be overwritten to change
#     # its behaviour if the specific transition needs
#     # it
#     def _process_frame(
#         self,
#         frame_a: Union['moderngl.Texture', 'np.ndarray'],
#         frame_b: Union['moderngl.Texture', 'np.ndarray'],
#         t_progress: float,
#     ):
#         # TODO: Maybe this can be placed in the general
#         # class if it doesn't change
#         frame = VideoFrame.from_ndarray(
#             array = self._get_process_frame_array(
#                 frame_a = frame_a,
#                 frame_b = frame_b,
#                 t_progress = t_progress
#             # TODO: Force this with the 'yta-numpy' utils to 'np.uint8'
#             ).astype(np.uint8),
#             format = 'rgb24'
#         )

#         # frame = VideoFrame.from_ndarray(
#         #     array = texture_to_frame(
#         #         texture = self._transition_node.process(
#         #             input_a = frame_a,
#         #             input_b = frame_b,
#         #             progress = t_progress
#         #         ),
#         #         do_include_alpha = False
#         #     ).astype(np.uint8),
#         #     format = 'rgb24'
#         # )

#         # The 'pts' and that is not important here but...
#         frame.pts = None
#         frame.time_base = Fraction(1, 60)

#         return frame
    
#     # TODO: Overwrite this method to make it custom (?)
#     def _get_process_frame_array(
#         self,
#         frame_a: Union['moderngl.Texture', 'np.ndarray'],
#         frame_b: Union['moderngl.Texture', 'np.ndarray'],
#         t_progress: float,
#     ) -> np.ndarray:
#         """
#         Get the numpy array that will be used to build the
#         pyav VideoFrame that will be returned.
#         """
#         return self._transition_node.process(
#             input_a = frame_a,
#             input_b = frame_b,
#             progress = t_progress
#         )
    
#     # TODO: Add 'save_frame_as'

#     def save_as(
#         self,
#         output_filename: str,
#         video_size: tuple[int, int] = None,
#         video_fps: float = None,
#         video_codec: str = None,
#         video_pixel_format: str = None,
#         audio_codec: str = None,
#         audio_sample_rate: int = None,
#         audio_layout: str = None,
#         audio_format: str = None,
#         do_apply_video_filters: bool = True,
#         do_apply_audio_filters: bool = True
#     ) -> str:
#         """
#         Save the file as 'output_filename'.

#         This method is useful if you want to apply
#         some filter and then save the video with
#         those filters applied into a new one, maybe
#         with a new pixel format and/or code. You can
#         prepare alpha transitions, etc.
#         """
#         video_size = (
#             getattr(self, 'size', Settings.DEFAULT_VIDEO_SIZE.value)
#             if video_size is None else
#             video_size
#         )

#         video_fps = (
#             getattr(self, 'fps', Settings.DEFAULT_VIDEO_FPS.value)
#             if video_fps is None else
#             video_fps
#         )

#         video_codec = (
#             getattr(self, 'codec_name', Settings.DEFAULT_VIDEO_CODEC.value)
#             if video_codec is None else
#             video_codec
#         )

#         video_pixel_format = (
#             getattr(self, 'pixel_format', Settings.DEFAULT_PIXEL_FORMAT.value)
#             if video_pixel_format is None else
#             video_pixel_format
#         )

#         audio_codec = (
#             getattr(self, 'audio_codec_name', Settings.DEFAULT_AUDIO_CODEC.value)
#             if audio_codec is None else
#             audio_codec
#         )

#         audio_sample_rate = (
#             getattr(self, 'audio_fps', Settings.DEFAULT_AUDIO_FPS.value)
#             if audio_sample_rate is None else
#             audio_sample_rate
#         )

#         audio_layout = (
#             getattr(self, 'audio_layout', Settings.DEFAULT_AUDIO_LAYOUT.value)
#             if audio_layout is None else
#             audio_layout
#         )

#         audio_format = (
#             getattr(self, 'audio_format', Settings.DEFAULT_AUDIO_FORMAT.value)
#             if audio_format is None else
#             audio_format
#         )

#         writer = VideoWriter(output_filename)

#         # TODO: This has to be dynamic according to the
#         # video we are writing (?)
#         writer.set_video_stream(
#             codec_name = video_codec,
#             fps = video_fps,
#             size = video_size,
#             pixel_format = video_pixel_format
#         )
        
#         writer.set_audio_stream(
#             codec_name = audio_codec,
#             fps = audio_sample_rate,
#             layout = audio_layout,
#             format = audio_format
#         )

#         # TODO: Maybe we need to reformat or something
#         # if some of the values changed, such as fps,
#         # audio sample rate, etc. (?)

#         time_base = fps_to_time_base(video_fps)
#         audio_time_base = fps_to_time_base(audio_sample_rate)

#         for t in get_ts(0, self._t_end, video_fps):
#             frame = self.get_video_frame_at(
#                 t = t,
#                 do_apply_filters = do_apply_video_filters
#             )

#             # TODO: What if 'frame' is None (?)
#             if frame is None:
#                 print(f'   [ERROR] Frame not found at t:{float(t)}')
#                 continue

#             writer.mux_video_frame(
#                 frame = frame
#             )

#             frame.time_base = time_base
#             frame.pts = T(t, time_base).truncated_pts

#             # TODO: Make this work
#             # audio_pts = 0
#             # for audio_frame in self.get_audio_frames_at(
#             #     t = t,
#             #     video_fps = video_fps,
#             #     do_apply_filters = do_apply_audio_filters
#             # ):
#             #     # TODO: 'audio_frame' could be None or []
#             #     # here if no audio channel
#             #     if audio_frame is None:
#             #         # TODO: Generate silence audio to cover the
#             #         # whole video frame (?)
#             #         pass
                
#             #     # We need to adjust our output elements to be
#             #     # consecutive and with the right values
#             #     # TODO: We are using int() for fps but its float...
#             #     audio_frame.time_base = audio_time_base
#             #     audio_frame.pts = audio_pts

#             #     # We increment for the next iteration
#             #     audio_pts += audio_frame.samples

#             #     writer.mux_audio_frame(audio_frame)

#         writer.mux_video_frame(None)
#         writer.mux_audio_frame(None)
#         writer.output.close()

#         return output_filename
    
# from yta_editor_nodes.processor.video.transitions import CrossfadeTransitionProcessor
    
# # TODO: This should be moved, maybe, to another file
# class CrossfadeTransitionMedia(_TransitionMedia):
#     """
#     A simple crossfade transition.
#     """

#     def __init__(
#         self,
#         clip_a: _VideoMedia,
#         clip_b: _VideoMedia,
#         duration: float,
#         rate_function: RateFunction = RateFunction.LINEAR,
#     ):
#         super().__init__(
#             clip_a = clip_a,
#             clip_b = clip_b,
#             duration = duration,
#             rate_function = rate_function
#         )
        
#         self._transition_node: CrossfadeTransitionProcessor = CrossfadeTransitionProcessor(
#             do_use_gpu = True,
#             output_size = (1920, 1080)
#         )


"""
Audio sources, that are the source from
where we will obtain the data to offer
as audio in our editor.

These sources will be used by other
classes to access to the frames but 
improve the functionality and simplify
it.
"""
from yta_editor.sources.abstract import _AudioSource
from yta_video_pyav.reader.filter.dataclass import GraphFilter
from yta_video_pyav.reader import AudioReader
from yta_validation.parameter import ParameterValidator
from av.audio.frame import AudioFrame
from quicktions import Fraction
from typing import Union

import numpy as np


class AudioFileSource(_AudioSource):
    """
    Class to represent an audio, read from an audio
    file, as an audio media source.

    You can apply effects directly to the source via
    ffmpeg filters.
    """

    @property
    def copy(
        self
    ) -> 'AudioFileSource':
        """
        Get a copy of this instance.
        """
        return AudioFileSource(
            filename = self.filename,
            audio_filters = self._audio_filters
        )

    @property
    def duration(
        self
    ) -> Fraction:
        """
        The duration of the audio.
        """
        return self.reader.audio_duration
    
    @property
    def audio_fps(
        self
    ) -> Union[int, None]:
        """
        The frames per second of the audio.
        """
        return self.reader.audio_fps
    
    @property
    def audio_codec_name(
        self
    ) -> Union[str, None]:
        """
        The name of the audio codec.
        """
        return self.reader.audio_codec_name
    
    @property
    def audio_layout(
        self
    ) -> Union[str, None]:
        """
        The audio layout.
        """
        return self.reader.audio_layout
    
    @property
    def audio_format(
        self
    ) -> Union[str, None]:
        """
        The audio format.
        """
        return self.reader.audio_format
    
    @property
    def audio_time_base(
        self
    ) -> Union[Fraction, None]:
        """
        The time base of the audio.
        """
        return self.reader.audio_time_base
    
    def __init__(
        self,
        filename: str,
        audio_filters: list[GraphFilter] = []
    ):
        self.filename: str = filename
        """
        The filename of the original audio.
        """
        self.reader: AudioReader = AudioReader(
            filename = self.filename,
            audio_filters = audio_filters
        )
        """
        The pyav audio reader.
        """
        self._audio_filters = audio_filters
        """
        The filters we want to apply to each audio
        frame.
        """

    def add_audio_filter(
        self,
        filter: GraphFilter
    ) -> 'AudioFileSource':
        """
        Add an audio filter to the list of filters
        to apply.
        """
        ParameterValidator.validate_mandatory_instance_of('filter', filter, GraphFilter)

        # TODO: Maybe handle repeated ones (?)
        self._audio_filters.append(filter)

        return self

    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_apply_filters: bool = True
    ):
        """
        Get the sequence of audio frames for a 
        given video 't' time moment, using the
        audio cache system.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).
        """
        for frame in self.reader.get_audio_frames_at(
            t = t,
            video_fps = video_fps,
            do_apply_filters = do_apply_filters
        ):
            yield frame

# TODO: This 'AudioNumpySource' class is
# very experimental, it needs refactor
# and I don't know if we will use it...
class AudioNumpySource(_AudioSource):
    """
    Class to represent an audio, made from a
    numpy array, as an audio media source.

    This source is static. The same audio
    frame will be returned always.
    """

    @property
    def copy(
        self
    ) -> 'AudioNumpySource':
        """
        Get a copy of this instance.
        """
        return AudioNumpySource(self._array)

    @property
    def duration(
        self
    ):
        """
        The duration of the source.
        """
        # TODO: Should I return something like 999 (?)
        return None

    # TODO: Put some information about the
    # shape we need to pass, and also create
    # a 'duration' property
    @property
    def frame(
        self
    ) -> AudioFrame:
        """
        The frame that must be played.
        """
        # TODO: What 'format' do we use? I think we
        # need to inspect the array to auto detect
        # it

        # return {
        #     's16': np.int16,
        #     'flt': np.float32,
        #     'fltp': np.float32
        # }.get(audio_format, None)
    
        # By now I'm forcing to this
        return AudioFrame.from_ndarray(
            array = self._array,
            format = 'fltp',
            layout = 'stereo'
        )

    def __init__(
        self,
        array: np.ndarray,
        # TODO: I think I need more information
        # to know how to read it
        # sample_rate (?)
    ):
        self._array: np.ndarray = array
        """
        The array of information that will be
        used to make the frame that will be
        played its whole duration.
        """
        # TODO: We should autodetect format,
        # layout and 'duration' and 'sample_rate'
        # that we will call 'audio_fps' through
        # a property

    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        # TODO: I need it but it is not used
        do_apply_filters: bool = False
    ):
        """
        Get the sequence of audio frames for a 
        given video 't' time moment.

        This is useful when we want to write a
        video frame with its audio, so we obtain
        all the audio frames associated to it
        (remember that a video frame is associated
        with more than 1 audio frame).

        As this is an audio from a static numpy
        array, the duration must fit 1/video_fps.
        """
        # TODO: We need to concatenate the audio
        # to make it fit the 1/video_fps duration
        # or maybe this class is unnecessary...
        yield self.frame


"""
Module to include the transitions that are made
with the OpenGL engine.

TODO: This module has to be rebuilt
"""
from yta_editor.tracks.media.video import VideoTimed
from yta_editor.media.video import _VideoMedia
from yta_video_opengl.nodes.video.transitions.crossfade import CrossfadeTransitionNode, DistortedCrossfadeTransitionNode
from yta_video_opengl.nodes.video.transitions.slide import SlideTransitionNode
from yta_video_opengl.nodes.video.transitions.alphapedia import AlphaPediaMaskTransitionNode
from yta_video_opengl.nodes.video.transitions import CircleOpeningTransitionNode, CircleClosingTransitionNode, BarsFallingTransitionNode
from yta_video_opengl.utils import texture_to_frame
from yta_video_opengl.context import OpenGLContext
from yta_video_frame_time.t_fraction import get_ts, fps_to_time_base, T
from yta_math.rate_functions.rate_function import RateFunction
from yta_video_pyav.writer import VideoWriter
from yta_video_pyav.settings import Settings
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from quicktions import Fraction
from typing import Union

import numpy as np


class _OpenglTransitionMedia:
    """
    *For internal use only*

    This class must be inherited by the transitions
    that have the specific implementation.

    Class to wrap a transition clip, that is a
    clip in which 2 clips are joined with another
    transition clip.

    This class, by now, is only handling clips that
    are played together during the 'transition'
    period of time. No frozen frames, no diferent
    time moments. When the transition is executed
    both videos are being played at the same time.
    """

    @property
    def _t_start(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The moment in which the transition clip starts
        being played, that will be in some moment of
        the first clip, and it should be also the 
        moment in which the second clip starts being
        played (with the corresponding transition
        blending).
        """
        # TODO: This value is only valid for the transition
        # applied in the middle with 50%-50% strategy
        return self._clip_a.duration - self.duration
    
    @property
    def _t_end(
        self
    ) -> Fraction:
        """
        *For internal use only*

        The moment in which the whole clip built with
        the transition in the middle must stop being
        played, so the second clip has finished.
        """
        return self._clip_b.end

    def __init__(
        self,
        # TODO: This '_MediaTimed' has to be renamed 
        # because here it is not on a track but the class
        # we need is this one
        # TODO: Maybe it has to be just a Media, not OnTrack
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        ParameterValidator.validate_subclass_of('clip_a', clip_a, _VideoMedia)
        ParameterValidator.validate_subclass_of('clip_b', clip_b, _VideoMedia)

        if (
            clip_a.duration < duration or
            clip_b.duration < duration
        ):
            raise Exception('The duration of one of the clips is smaller than the transition duration requested.')

        # TODO: Apply Union[AlphaBlendTransition, ...]
        self.duration: float = duration
        """
        The duration of the transition.
        """
        self.rate_function: RateFunction = RateFunction.to_enum(rate_function)
        """
        The rate function to be applied in the transition
        progress to handle its speed.
        """

        self._clip_a: VideoTimed = VideoTimed(
            media = clip_a,
            start = 0
        )
        """
        The first clip to join in with the transition.
        """
        self._clip_b: VideoTimed = VideoTimed(
            media = clip_b,
            start = self._t_start
        )
        """
        The second clip to join in with the transition.
        """

        # TODO: The child class must implement the
        # 'self.opengl_node' variable
        # self.opengl_node: '_OpenglNodeBase' = CrossfadeTransitionNode(
        #     opengl_context = OpenGLContext().context,
        #     # TODO: Do not hardcode, please
        #     size = (1920, 1080),
        # )
    
    def get_transition_progress_at(
        self,
        t: float
    ) -> float:
        """
        Get the transition progress value, that will
        be in the [0.0, 1.0] range, according to the
        `t` time moment provided and the `rate_function`
        that must be applied in this transition.

        This method should be called only when the
        transition has to be applied, so the `t` is
        a time moment in which the two frames are 
        being played together.
        """

        """
        By now, the transition we are applying is
        as simple as being in the middle of the 2
        clips provided and lasting the 'duration'
        provided, being executed with a linear t.
        """
        # The transition can start in a time moment
        # different than 0 so we recalculate it
        t -= self._t_start

        return (
            0.0
            if t < 0 else
            1.0
            if t > self.duration else
            self.rate_function.get_n_value(
                n = max(
                    0.0,
                    min(
                        1.0,
                        t / self.duration
                    )
                )
            )
        )
    
    def get_video_frame_at(
        self,
        t: Union[int, float, Fraction],
        do_apply_filters: bool = False
    ) -> VideoFrame:
        """
        The `t` time moment provided must be a global
        time value pointing to a moment in the general
        timeline. That value will be transformed to the
        internal `t` value to get the frame.
        """
        # Test with OpenGL
        frame_a = self._clip_a.get_video_frame_at(t)
        frame_b = self._clip_b.get_video_frame_at(t)

        if (
            frame_a is None and
            frame_b is None
        ):
            print('Both None, wtf')
            #return None

        """
        There is a moment in which we are not in
        the transition section but on the clips
        so we just need to return the clip frames.
        """
        if frame_b is None:
            # Not transition time, just one clip
            out_frame = frame_a
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame
        
        if frame_a is None:
            # Not transition time, just one clip
            out_frame = frame_b
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame
        
        # Render loop
        frame_a = frame_a.to_ndarray(format = 'rgb24').astype(np.float32)
        frame_b = frame_b.to_ndarray(format = 'rgb24').astype(np.float32)

        # TODO: Do we need the alpha (not rgb24?) maybe
        # to mix with something else
        return self._process_frame(
            frame_a = frame_a,
            frame_b = frame_b,
            t_progress = self.get_transition_progress_at(t)
        )
    
    # TODO: This method can be overwritten to change
    # its behaviour if the specific transition needs
    # it
    def _process_frame(
        self,
        # numpy array or texture
        frame_a: any,
        # numpy array or texture
        frame_b: any,
        t_progress: float,
    ):
        # TODO: Maybe this can be placed in the general
        # class if it doesn't change
        frame = VideoFrame.from_ndarray(
            array = texture_to_frame(
                texture = self.opengl_node.process(
                    input_a = frame_a,
                    input_b = frame_b,
                    progress = t_progress
                ),
                do_include_alpha = False
            ).astype(np.uint8),
            format = 'rgb24'
        )
        # The 'pts' and that is not important here but...
        frame.pts = None
        frame.time_base = Fraction(1, 60)

        return frame
    
    # TODO: Add 'save_frame_as'

    def save_as(
        self,
        output_filename: str,
        video_size: tuple[int, int] = None,
        video_fps: float = None,
        video_codec: str = None,
        video_pixel_format: str = None,
        audio_codec: str = None,
        audio_sample_rate: int = None,
        audio_layout: str = None,
        audio_format: str = None,
        do_apply_video_filters: bool = True,
        do_apply_audio_filters: bool = True
    ) -> str:
        """
        Save the file as 'output_filename'.

        This method is useful if you want to apply
        some filter and then save the video with
        those filters applied into a new one, maybe
        with a new pixel format and/or code. You can
        prepare alpha transitions, etc.
        """
        video_size = (
            getattr(self, 'size', Settings.DEFAULT_VIDEO_SIZE.value)
            if video_size is None else
            video_size
        )

        video_fps = (
            getattr(self, 'fps', Settings.DEFAULT_VIDEO_FPS.value)
            if video_fps is None else
            video_fps
        )

        video_codec = (
            getattr(self, 'codec_name', Settings.DEFAULT_VIDEO_CODEC.value)
            if video_codec is None else
            video_codec
        )

        video_pixel_format = (
            getattr(self, 'pixel_format', Settings.DEFAULT_PIXEL_FORMAT.value)
            if video_pixel_format is None else
            video_pixel_format
        )

        audio_codec = (
            getattr(self, 'audio_codec_name', Settings.DEFAULT_AUDIO_CODEC.value)
            if audio_codec is None else
            audio_codec
        )

        audio_sample_rate = (
            getattr(self, 'audio_fps', Settings.DEFAULT_AUDIO_FPS.value)
            if audio_sample_rate is None else
            audio_sample_rate
        )

        audio_layout = (
            getattr(self, 'audio_layout', Settings.DEFAULT_AUDIO_LAYOUT.value)
            if audio_layout is None else
            audio_layout
        )

        audio_format = (
            getattr(self, 'audio_format', Settings.DEFAULT_AUDIO_FORMAT.value)
            if audio_format is None else
            audio_format
        )

        writer = VideoWriter(output_filename)

        # TODO: This has to be dynamic according to the
        # video we are writing (?)
        writer.set_video_stream(
            codec_name = video_codec,
            fps = video_fps,
            size = video_size,
            pixel_format = video_pixel_format
        )
        
        writer.set_audio_stream(
            codec_name = audio_codec,
            fps = audio_sample_rate,
            layout = audio_layout,
            format = audio_format
        )

        # TODO: Maybe we need to reformat or something
        # if some of the values changed, such as fps,
        # audio sample rate, etc. (?)

        time_base = fps_to_time_base(video_fps)
        audio_time_base = fps_to_time_base(audio_sample_rate)

        for t in get_ts(0, self._t_end, video_fps):
            frame = self.get_video_frame_at(
                t = t,
                do_apply_filters = do_apply_video_filters
            )

            # TODO: What if 'frame' is None (?)
            if frame is None:
                print(f'   [ERROR] Frame not found at t:{float(t)}')
                continue

            writer.mux_video_frame(
                frame = frame
            )

            frame.time_base = time_base
            frame.pts = T(t, time_base).truncated_pts

            # TODO: Make this work
            # audio_pts = 0
            # for audio_frame in self.get_audio_frames_at(
            #     t = t,
            #     video_fps = video_fps,
            #     do_apply_filters = do_apply_audio_filters
            # ):
            #     # TODO: 'audio_frame' could be None or []
            #     # here if no audio channel
            #     if audio_frame is None:
            #         # TODO: Generate silence audio to cover the
            #         # whole video frame (?)
            #         pass
                
            #     # We need to adjust our output elements to be
            #     # consecutive and with the right values
            #     # TODO: We are using int() for fps but its float...
            #     audio_frame.time_base = audio_time_base
            #     audio_frame.pts = audio_pts

            #     # We increment for the next iteration
            #     audio_pts += audio_frame.samples

            #     writer.mux_audio_frame(audio_frame)

        writer.mux_video_frame(None)
        writer.mux_audio_frame(None)
        writer.output.close()

        return output_filename

class OpenGLCrossfadeTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a crossfade
    in between the frames.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function
        )

        self.opengl_node: '_OpenglNodeBase' = CrossfadeTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

class OpenGLDistortedCrossfadeTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a distorted
    crossfade in between the frames.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
        intensity: float = 0.1,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )

        self.opengl_node: '_OpenglNodeBase' = DistortedCrossfadeTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
            intensity = intensity
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

class OpenGLSlideTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a slide 
    between the frames.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
        # TODO: This must be an Enum and handled also in
        # the 'yta_video_opengl' library
        direction: int = 0,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )

        # TODO: Do I need it here or just in the node
        # instance (?)
        self._direction: int = direction
        """
        The direction of the slide.
        """

        self.opengl_node: '_OpenglNodeBase' = SlideTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
            direction = direction
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

class OpenGLCircleOpeningTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a circle
    growing from the middle of the video.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )

        self.opengl_node: '_OpenglNodeBase' = CircleOpeningTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
            resolution = (1920, 1080),
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

class OpenGLCircleClosingTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a circle
    disappearing in the middle of the screen.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )

        self.opengl_node: '_OpenglNodeBase' = CircleClosingTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
            resolution = (1920, 1080),
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

class OpenGLAlphaPediaTransitionMedia(_OpenglTransitionMedia):
    """
    A transition that joins 2 videos by applying
    another video as a mask, that is specifically
    built for this purpose and stored in the
    AlphaPediaYT channel.

    The transition must be a black background that
    changes to white, where black will be the 
    previous frame and the white will be replaced
    by the next frame.

    These videos are made with black and white
    colors that are transformed into an alpha
    channel according to the presence of that 
    white color and becoming a mask.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        # TODO: Maybe url instead (?)
        clip_mask: _VideoMedia,
        # The video has its own duration so we can
        # use it (if possible)
        duration: Union[float, None],
        rate_function: RateFunction = RateFunction.LINEAR,
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )
        
        # TODO: I think we don't need it to be a
        # VideoTimed. As a Media is ok to get
        # the frame
        self._mask_clip = VideoTimed(
            media = clip_mask,
            start = 0
        )

        self.opengl_node: '_OpenglNodeBase' = AlphaPediaMaskTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

    def get_video_frame_at(
        self,
        t: Union[int, float, Fraction],
        do_apply_filters: bool = False
    ) -> VideoFrame:
        """
        The `t` time moment provided must be a global
        time value pointing to a moment in the general
        timeline. That value will be transformed to the
        internal `t` value to get the frame.
        """
        # Test with OpenGL
        frame_a = self._clip_a.get_video_frame_at(t)
        frame_b = self._clip_b.get_video_frame_at(t)

        if (
            frame_a is None and
            frame_b is None
        ):
            print('Both None, wtf')
            #return None

        """
        There is a moment in which we are not in
        the transition section but on the clips
        so we just need to return the clip frames.
        """
        if frame_b is None:
            # Not transition time, just one clip
            out_frame = frame_a
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame
        
        if frame_a is None:
            # Not transition time, just one clip
            out_frame = frame_b
            out_frame.pts = None
            out_frame.time_base = Fraction(1, 60)

            return out_frame
        
        frame_a = frame_a.to_ndarray(format = 'rgb24').astype(np.float32)
        frame_b = frame_b.to_ndarray(format = 'rgb24').astype(np.float32)

        """
        We will obtain the mask video frame that we need to
        send to the OpenGL transition handler class to apply
        in the provided 't' time moment, according to its
        duration, and also will use it to process the
        transition frame alpha blending.
        """
        t_progress = self.get_transition_progress_at(t)
        frame_mask = self._mask_clip.get_video_frame_at(t_progress * self._mask_clip.duration)
        # TODO: Could this frame be None (?)
        frame_mask = frame_mask.to_ndarray(format = 'rgb24').astype(np.float32)

        # TODO: Do we need the alpha (not rgb24?) maybe
        # to mix with something else
        return self._process_frame(
            frame_a = frame_a,
            frame_b = frame_b,
            frame_mask = frame_mask,
            t_progress = t_progress
        )

    def _process_frame(
        self,
        # numpy array or texture
        frame_a: any,
        # numpy array or texture
        frame_b: any,
        # numpy array or texture
        frame_mask: any,
        t_progress: float,
    ):
        frame = VideoFrame.from_ndarray(
            array = texture_to_frame(
                texture = self.opengl_node.process(
                    input_a = frame_a,
                    input_b = frame_b,
                    input_mask = frame_mask,
                    progress = t_progress
                ),
                do_include_alpha = False
            ).astype(np.uint8),
            format = 'rgb24'
        )
        # The 'pts' and that is not important here but...
        frame.pts = None
        frame.time_base = Fraction(1, 60)

        return frame

class OpenGLBarsFallingTransitionMedia(_OpenglTransitionMedia):
    """
    A transition made with OpenGL that is a set
    of bars falling.
    """

    def __init__(
        self,
        clip_a: _VideoMedia,
        clip_b: _VideoMedia,
        duration: float,
        rate_function: RateFunction = RateFunction.LINEAR,
        bars: int = 30,
        amplitude: float = 2.0,
        noise: float = 0.1,
        frequency: float = 0.5,
        drip_scale: float = 0.5
    ):
        super().__init__(
            clip_a = clip_a,
            clip_b = clip_b,
            duration = duration,
            rate_function = rate_function,
        )

        self.opengl_node: '_OpenglNodeBase' = BarsFallingTransitionNode(
            opengl_context = OpenGLContext().context,
            # TODO: Do not hardcode, please
            size = (1920, 1080),
            bars = bars,
            amplitude = amplitude,
            noise = noise,
            frequency = frequency,
            dripScale = drip_scale
        )
        """
        The OpenGL node that is actually the transition
        processor.
        """

# TODO: Build more transitions
    

"""
Davinci Resolve has these 3 types of transition
strategies:
- Start on Cut → La transición empieza justo en el corte, extendiéndose hacia adelante.
- Center on Cut → La transición se centra en el corte (la mitad sobre el primer clip y la mitad sobre el segundo).
- End on Cut → La transición termina justo en el corte, extendiéndose hacia atrás.
"""



from yta_editor.tracks.items.gap import GapTrackItem
from yta_editor.utils.silence import generate_silent_frames
from yta_editor.tracks.items.transition import TransitionMode, TransitionTrackItem
from yta_video_frame_time.t_fraction import check_values_are_same
from yta_time_interval import TIME_INTERVAL_SYSTEM_LIMITS
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from quicktions import Fraction
from typing import Union
from abc import ABC
from dataclasses import dataclass

import bisect


TrackItemType = Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem']
"""
The type that includes all the specific track item
types we can handle.
"""

@dataclass
class _TrackItemFound:
    """
    Dataclass just to hold the items we found in the
    items list, including the item but also the index.

    This dataclass is not for being stored within the
    `_TrackItems` list but to be returned when looking
    for specific items in the list.
    """

    def __init__(
        self,
        item: TrackItemType,
        index: int
    ):
        self.item: TrackItemType = item
        """
        The item found.
        """
        self.index: int = index
        """
        The index of the item found in the `items` list.
        """

@dataclass
class _TrackItems:
    """
    The list of items that exist in the track.

    This class will implement optimized ways of looking
    for items based on the time moments they have. The
    items inside will be always ordered by the `t_start`
    time moment.
    """

    @property
    def is_empty(
        self
    ) -> bool:
        """
        Boolean flag that indicates if the list is empty or
        not.
        """
        return len(self.items) == 0

    def __init__(
        self,
        track: Union['VideoTrack', 'AudioTrack']
    ):
        self.items: list[TrackItemType] = []
        """
        The list of items, that will be always initialized
        with a gap that lasts the whole duration.
        """
        # TODO: I don't know exactly how to handle this... What
        # if one of the items provided later doesn't belong to
        # this track (?)
        self._track: Union['VideoTrack', 'AudioTrack'] = track
        """
        The track of these items belong to.
        """

        self._fullfil_if_empty()

    """
    Internal but necessary and very important methods below.
    """

    def _recalculate_all(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        This method should be called when we add or remove
        any item from the track, so the track remains
        consistent, with the items ordered and with
        everything valid inside.
        """
        print('[OK] Ordering items')
        self._order_items()
        print('[OK] Relinking items')
        self._relink_items()
        print('[OK] Rebuilding gaps')
        self._rebuild_gaps()
        print('[OK] Merging consecutive gaps')
        self._merge_consecutive_gaps()
        print('[OK] Validating all the items inside')
        self._validate_items()
        print('[OK] Validate the transitions (if existing)')
        self._validate_transitions()

    def _fullfil_if_empty(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Fulfill the list with a gap that lasts the whole
        duration if the list is completely empty, as we
        always need to have one item inside.
        """
        if self.is_empty:
            self.items.append(
                GapTrackItem(
                    track = self._track,
                    t_start = 0,
                    duration = TIME_INTERVAL_SYSTEM_LIMITS[1],
                    item_in = None,
                    item_out = None
                )
            )

        return self
    
    def _validate_items(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Check that the `items` list is valid and all the items
        are consecutive, with no empty gap in between and with
        no overlappings.
        """
        # TODO: The list must be ordered
        for previous_item, current_item in zip(self.items, self.items[1:]):
            if previous_item.t_end != current_item.t_start:
                raise Exception(f'There is an item with "t_end={str(float(previous_item.t_end))}" and the next one with "t_start={str(float(current_item.t_start))}".')

        return self
    
    def _order_items(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Order the `items` list and ensures that all the items
        are ordered by the `t_start` field. This method will
        updated the `self.items` list.
        """
        self.items.sort(key = lambda item: item.t_start)

        return self
    
    def _relink_items(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Reset all the links in between the items that are in
        the list. One item will be the `item_in` of the next
        one, which will also be the `item_out` of that item
        that is inmediately before.
        """
        # 1. Order the items again based on the `t_start` field
        # self._reorder_items()

        # 2. Reset links
        prev = None
        for item in self.items:
            item.item_in = prev

            if prev:
                prev.item_out = item

            prev = item

        if prev:
            prev.item_out = None

        # 3. Regenerate gaps
        # self._rebuild_gaps()

        # 4. Validate transitions
        # self._validate_transitions()

    def _merge_consecutive_gaps(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        This method will merge the consecutive gaps existing
        in the items list, to combine them as a single but 
        longer one.
        """
        if self.is_empty:
            return

        merged = []
        current = self.items[0]

        for item in self.items[1:]:
            if (
                is_gap_track_item(current) and
                is_gap_track_item(item)
            ):
                # TODO: This should always happen. If it doesn't happen
                # is because we did something wrong...
                if item.t_start <= current.t_end:
                    # Combine gaps in one
                    current.shift_end_to(max(current.t_end, item.t_end))
                    # Point the 'item_out' to the next item's out
                    current.item_out = item.item_out
                    continue

            # No more consecutives to combine, just add
            merged.append(current)
            current = item

        # Add the last one pending and replace
        merged.append(current)
        self.items = merged

        return self

    def _rebuild_gaps(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Rebuild the gaps according to a change that has happened
        in some items, modifying the size of the gaps.
        """
        # TODO: Modify the time interval of any gap that doesn't
        # match the time moments of the track (that doesn't end
        # when the next item starts, or that ends after another
        # item has started)
        # No tiene sentido si hay menos de 2 items
        if len(self.items) < 2:
            return self
        
        # TODO: Should we build gaps in between elements that have
        # been shifted or something, leaving a time interval gap 
        # between the `t_start` of the first one and the `t_end`
        # of the second one (?)

        for i, item in enumerate(self.items):
            if not PythonValidator.is_instance_of(item, GapTrackItem):
                continue

            previous_item = (
                self.items[i - 1]
                if (i - 1) >= 0 else
                None
            )

            next_item = (
                self.items[i + 1]
                if (i + 1) < len(self.items) else
                None
            )

            # Set the time interval that is the right one
            if (
                previous_item is not None and
                item.t_start != previous_item.t_end
            ):
                # TODO: Use a 'item._set_t_start()' instead (?)
                #item._time_interval.with_start_and_end(previous_item.t_end, item.t_end)
                item._time_interval.t_start = previous_item.t_end

            if (
                next_item is not None and
                item.t_end != next_item.t_start
            ):
                # TODO: Use a 'item._set_t_end()' instead (?)
                #item._time_interval.with_start_and_end(item.t_start, next_item.t_start)
                item._time_interval.t_end = next_item.t_start

        # If we had 2 consecutive gaps, this method is fixing
        # the time gap if existing, but not combining them. 
        # Use the '_merge_consecutive_gaps' to do that.
        
        return self

    def _validate_transitions(
        self
    ) -> '_TrackItems':
        """
        *For internal use only*

        Check and validate the transitions that exist before any
        change to remove the ones that are not available now 
        according to the new changes.
        """
        # TODO: Maybe we don't need to raise exception but to
        # fix the time intervals of these transitions
        for i, item in enumerate(self.items):
            if not PythonValidator.is_instance_of(item, 'TransitionTrackItem'):
                continue

            if (
                i == 0 or
                i == len(self.items) - 1
            ):
                # I think this will never happen based on the
                # way the system is built
                raise Exception('Transition item cannot be the first or last item.')

            item_in = self.items[i - 1]
            item_out = self.items[i + 1]

            if item.item_in is not item_in:
                print(f'Transition item_in mismatch: expected index {i-1}, but transition.item_in is {item.item_in}.')
                # TODO: Remove any unexpected transition link
                item.item_in.transition_out = None
                self.remove_item(item)
                #raise Exception(f'Transition item_in mismatch: expected index {i-1}, but transition.item_in is {item.item_in}.')

            if item.item_out is not item_out:
                raise Exception(f'Transition item_out mismatch: expected index {i+1}, but transition.item_out is {item.item_out}.')

            if item_in.t_end != item.t_start:
                raise Exception(f'Transition start mismatch: item_in.t_end={float(item_in.t_end)} != transition.t_start={float(item.t_start)}')

            if item.t_end != item_out.t_start:
                raise Exception(f'Transition end mismatch: transition.t_end={float(item.t_end)} != item_out.t_start={float(item_out.t_start)}')
        
        return self
    
    def _shift_next_elements(
        self,
        item: TrackItemType,
        delta: Union[int, float, Fraction],
        do_include_item: bool = False
    ):
        """
        *For internal use only*

        Update the next elements of the given `item` by applying
        the `delta` time amount provided, that will make their
        time intervals be updated with that value. If the 
        `do_include_item` boolean parameter is True, the `item`
        itself will be also shifted (useful when you have deleted
        one item and want to shift the next ones).

        This method must be called when we are adding an item 
        that was longer than expected and we need to move the
        next items.
        """
        print(f'shifting next elements delta={str(float(delta))}')

        item = (
            item
            if do_include_item else
            item.item_out
        )

        while item is not None:
            # We just need to move the time interval, but the connections
            # and everything are the same because we just inserted
            # something that makes this 'shift' necessary
            item._shift_by(delta)
            item = item.item_out

    def _replace_item(
        self,
        index: int,
        items: Union[TrackItemType, list[TrackItemType]]
    ) -> '_TrackItems':
        """
        *For internal use only*

        Replace the item at the `index` position with the item
        or items given as `items` parameter.

        If any element of the `items` array provided is None it
        will be omitted.
        """
        ParameterValidator.validate_int_between('index', index, 0, len(self.items), do_include_upper_limit = False)

        items = (
            [items]
            if not PythonValidator.is_list(items) else
            items
        )

        # Replace the item at t with the new composition
        items_to_insert = [
            item
            for item in items
            if item is not None
        ]

        if len(items_to_insert) == 0:
            raise Exception('The amount of "items" provided that are not None is 0.')

        # TODO: Validate that t_start and t_end are congruent
        # TODO: Validate that the previous and next items
        # are ok according to the timing of this item
        self.items[index:index + 1] = items_to_insert

        return self
    
    """
    Internal but necessary and very important methods above.
    """

    def reset(
        self
    ):
        """
        Reset the items list by emptying it and adding a
        single gap item lasting the whole track duration.

        (!) This method should be called carefully as it
        will remove everything built before.
        """
        self.items = []
        self._fullfil_if_empty()

    def get_item(
        self,
        item: TrackItemType
    ) -> Union['_TrackItemFound', None]:
        """
        Get the `item` provided from the items list, if
        existing, but as a `_TrackItemFound` including the
        index, or `None` if it doesn't exist on the list.
        """
        try:
            index = self.items.index(item)

            return _TrackItemFound(
                item = item,
                index = index
            )
        except:
            return None

    def get_item_at(
        self,
        t: Union[int, float, Fraction]
    ) -> Union['_TrackItemFound', None]:
        """
        Get the item that is placed at the given `t` global
        time moment in this track.

        We will have always at least one item at that position
        because we fill empty gaps with the special gap item.
        """
        # TODO: This should not happen as we should always
        # have at least a gap that fulfills the whole track
        # items list
        t_starts = [
            i.t_start
            for i in self.items
        ]

        # TODO: What if 0 (?)
        index = bisect.bisect_right(t_starts, t)

        # TODO: Why this? Is it necessary (?)
        # TODO: I had to do this because it was becoming -1 in
        # some situations
        index = (
            index - 1
            if index > 0 else
            index
        )

        return _TrackItemFound(
            item = self.items[index],
            index = index
        )
    
    def get_items_at(
        self,
        t_start: Union[int, float, Fraction],
        t_end: Union[int, float, Fraction]
    ) -> list['_TrackItemFound']:
        """
        Get all the items that will take place in the time
        interval `[t_start, t_end]`.
        """
        t_starts = [
            i.t_start
            for i in self.items
        ]

        # First element with t_start > t1
        index = bisect.bisect_right(t_starts, t_end)

        result = []

        # Go backwards until we have no overlapping
        i = index - 1
        while i >= 0:
            item = self.items[i]

            if item.t_end <= t_start:
                # No more overlapping
                break

            result.append(item)
            i -= 1

        # Go forwards (just to make sure it is well done)
        i = index
        while i < len(self.items):
            item = self.items[i]

            if item.t_start >= t_end:
                # No more overlapping
                break

            if item.t_end > t_start:
                result.append(item)

            i += 1

        return result
    
    def get_items_after(
        self,
        t: Union[int, float, Fraction]
    ) -> list['_TrackItemFound']:
        """
        Get all the items whose `t_start` time moment is
        equal or greater than the `t` time moment provided.
        """
        item_at_t = self.get_item_at(t)

        current_index = item_at_t.index + 1
        items = []
        for item in self.items[item_at_t.index:]:
            items.append(
                object = _TrackItemFound(
                    item = item,
                    index = current_index
                )
            )
            current_index += 1

        return items
    
    def get_items_before(
        self,
        t: Union[int, float, Fraction]
    ) -> list['_TrackItemFound']:
        """
        Get all the items whose `t_start` time moment is lower 
        than the `t` time moment provided.
        """
        item_at_t = self.get_item_at(t)

        current_index = item_at_t.index - 1
        items = []
        for item in self.items[:item_at_t.index]:
            items.append(
                object = _TrackItemFound(
                    item = item,
                    index = current_index
                )
            )
            current_index -= 1

        return items
    
    """
    Important methods below
    """
    
    def add_item(
        self,
        item: TrackItemType,
        # TODO: Create a 'mode' instead
        do_force_add: bool = True
    ) -> '_TrackItems':
        """
        Add the `item` provided to the list, that will be added
        in the corresponding place according to the `t_start` time
        moment of that `item`.
        """
        item_at_t = self.get_item_at(item.t_start)

        # TODO: Implement the 'do_force_add' to force adding
        # the item if the 't' is occupied

        """
        Opciones:
        - (1) Insertar VÍDEO en GAP que termina antes que el siguiente
        elemento:
            *Podemos insertar sin problema, detectando si hace falta gap
            en la izquierda y/o en la derecha y construyendo el triplete
            correspondiente
        - (2) Insertar VÍDEO en GAP que termina después de que el siguiente elemento
        empiece:
            *Tenemos que desplazar el inicio del elemento que está a la derecha
            hasta el momento en el que terminaría el que queremos insertar (y
            propagar el cambio a los siguientes), y colocar un gap a la izquierda
            para rellenar si hace falta.
        - (3) Insertar VÍDEO en VÍDEO. Comprobar la estrategia, y si está
        forzada, convertir el 't' recibido en el 't_end' del VÍDEO que ya hay, y
        operar como el punto (1) o (2).
        """

        if is_gap_track_item(item_at_t.item):
            """
            (1) and (2) are being handled here, inserting in
            a gap
            """
            if is_gap_track_item(item):
                # Item to add is a gap
                # TODO: Is this possible? I think we will not add
                # gaps but use them with the 'replace'...
                print('we are adding a gap')

            if item.t_end > item_at_t.item.t_end:
                """
                (2) Our item ends when the next item has started
                so we need to shift the next item a bit to create
                the space needed and then set the triplet
                """
                # Shift it (and propagate) the amount of time we need
                """
                The GAP is now shifting in an special way considering
                if it is the last one (special one) or not, so we can
                shift it correctly. By now the strategy I am applying
                is to shift everything that is on the right just as it
                is, so everything keeps being the same and has to be
                modified manually to actually change.
                """
                delta = item.t_end - item_at_t.item.t_end
                self._shift_next_elements(item_at_t.item, delta)

            """
            Our item ends before or at the t_start of the next
            one so we can add it easy by creating a triplet
            of left and right gaps if needed and replace the
            previous item with that new one.

            I'll give you a nice example to understand 
            everything better. Imagine a gap from t=3 to t=6
            in which we want to place a video of duration=2:
            - (1) Set at t=3 - The triplet will be:
                [None, Video[3, 5], Gap[5, 6]]
            - (2) Set at t=3.5 - The triplet will be:
                [Gap[3, 3.5], Video[3.5, 5.5], Gap[5.5, 6]
            - (3) Set at t=4 - The triplet will be:
                [Gap[3, 4], Video[4, 6], None]

            Now imagine that the gap we have is from t=3 to
            t=5, with the same video with duration=2:
            - (1) Set at t=3 - The triplet will be:
                [None, Video[3, 5], None]
            """

            left_gap_item = (
                GapTrackItem(
                    track = item._track,
                    t_start = item_at_t.item.t_start,
                    duration = item.t_start - item_at_t.item.t_start,
                    # I'll set it when I have the left and the right
                    item_in = None,
                    item_out = None,
                )
                if item.t_start > item_at_t.item.t_start else
                None
            )

            right_gap_item = (
                GapTrackItem(
                    track = item._track,
                    t_start = item.t_end,
                    duration = item_at_t.item.t_end - item.t_end,
                    # I'll set it when I have the left and the right
                    item_in = None,
                    item_out = None,
                )
                if item.t_end < item_at_t.item.t_end else
                None
            )

            # Keep the old connections
            old_gap_item_in = item_at_t.item.item_in
            old_gap_item_out = item_at_t.item.item_out
            # Unlink the original (old) gap
            item_at_t.item.unset_item_in()
            item_at_t.item.unset_item_out()

            # Keep the connections according to the changes
            if left_gap_item is not None:
                left_gap_item.set_item_out(item, True)
            else:
                item.set_item_in(old_gap_item_in)

            if right_gap_item is not None:
                right_gap_item.set_item_in(item, True)
            else:
                item.set_item_out(old_gap_item_out)

            # Replace the item at t with the new composition
            self._replace_item(
                index = item_at_t.index,
                items = [left_gap_item, item, right_gap_item]
            )
        else:
            """
            (3) We are trying to insert something in a place in
            which we already have a video (and not a gap) so we
            need to, instead of that, insert it at the 't_end' of
            that existing video, and then move the rest of the
            items that are on the right the duration of the item
            we added
            """
            if not do_force_add:
                raise Exception('No gap in that place, so sorry by now!')
            
            item_at_t_item_out = item_at_t.item.item_out
            # Change the 't_start.item' to the 'item_at_t.item.t_end'
            print(f'@@@ shifting item to {str(float(item_at_t.item.t_end))}')
            item.shift_to(item_at_t.item.t_end)
            item_at_t.item.set_item_out(item)
            item.set_item_out(item_at_t_item_out)

            # Shift all the elements to the right
            self._shift_next_elements(item, item.duration)

            self._replace_item(
                index = item_at_t.index,
                items = [item_at_t.item, item]
            )

        # TODO: We need to recalculate the links, etc.
        self._recalculate_all()

        return self
    
    def remove_item(
        self,
        item: TrackItemType,
        # TODO: Maybe create a 'mode' and use the mode instead
        do_create_gap_instead: bool = True
    ) -> '_TrackItems':
        """
        Remove the `item` from the list. The item, if 
        existing and different to a gap, will be transformed
        into a gap if the `do_create_gap_instead` boolean
        parameter is True, or it will be removed completely
        if it is False.

        The list will have one item less but will keep
        being ordered by the `t_start` time moment.
        """
        item: _TrackItemFound = self.get_item(item)

        # TODO: This should not happen I think...
        if item is None:
            raise Exception('The "item" was not found, what did you send bro?')
        
        # TODO: Refactor this
        if do_create_gap_instead:
            gap_item = GapTrackItem(
                track = item.item._track,
                t_start = item.item.t_start,
                duration = item.item.duration,
                # I link these 'in' and 'out' before
                item_in = None,
                item_out = None
            )

            if item.item.item_in is not None:
                gap_item.set_item_in(item.item)

            if item.item.item_out is not None:
                gap_item.set_item_out(item.item)

            # Is this necessary? We will lose the instance...
            item.item.unset_item_in(False)
            item.item.unset_item_out(False)
            
            self._replace_item(
                index = item.index,
                items = [gap_item]
            )
        else:
            """
            We have some special cases, such as being the only item in
            the track, which, even being strange, should be replaced
            by a gap. But also being the first or the last item of the
            track are special situations.
            """
            # TODO: Remove the item and shift the next ones
            previous_item = item.item.item_in
            next_item = item.item.item_out

            self.items.pop(item.index)

            # TODO: Review this below, I think I have to use 'False'
            # to avoid infinite loops
            if previous_item is not None:
                previous_item.set_item_out(next_item)

            if next_item is not None:
                next_item.set_item_in(previous_item)

            # Just for security
            item.item.item_in = None
            item.item.item_out = None

            self._shift_next_elements(next_item, -item.item.duration, True)

        self._recalculate_all()

        return self
    
    def remove_item_at(
        self,
        t: Union[int, float, Fraction],
        # TODO: Maybe create a 'mode' and use the mode instead
        do_create_gap_instead: bool = True
    ) -> '_TrackItems':
        """
        Remove the item from the list that corresponds to
        the `t` time moment, which means that is being
        displayed at that time (the `t` is in between the
        `[t_start, t_end]` time interval). The item, if 
        existing and different to a gap, will be transformed
        into a gap if the `do_create_gap_instead` boolean
        parameter is True, or it will be removed completely
        if it is False.

        The list will have one item less but will keep
        being ordered by the `t_start` time moment.
        """
        item_at_t = self.get_item_at(t)

        return self.remove_item(
            item = item_at_t.item,
            do_create_gap_instead = do_create_gap_instead
        )
    


    # TODO: Working on this one below
    def add_transition(
        self,
        # TODO: We need to know the way to handle this 'type'
        # and maybe all the properties of the transition
        item_in: TrackItemType,
        item_out: TrackItemType,
        type: any = 'random',
        mode: TransitionMode = TransitionMode.TRIM,
        duration: Union[int, float, Fraction] = 1
    ) -> '_TrackItems':
        """
        Create a transition between the `item_in` and
        `item_out` provided of the given `type`, if
        possible, and place it as a new item in between
        those 2 items that have to be consecutive and
        valid.
        """
        item_in: _TrackItemFound = self.get_item(item_in)
        item_out: _TrackItemFound = self.get_item(item_out)

        if item_in is None:
            raise Exception('The "item_in" does not exist in the track.')
        
        if item_out is None:
            raise Exception('The "item_out" does not exist in the track')
        
        # TODO: Validate that 'item_in.item' is a video item
        # TODO: Validate that 'item_out.item' is a video item
        
        if (item_in.index + 1) != item_out.index:
            raise Exception('The "item_in" and "item_out" provided are not consecutive.')

        mode = TransitionMode.to_enum(mode)

        if (
            item_in.item.duration <= duration or
            item_out.item.duration <= duration
        ):
            raise Exception(f'The items for the transition must last more time than the duration of the transition requested (duration={str(float(duration))}), and these are the durations of the items: "item_in duration={str(float(item_in.item.duration))}" and the "item_out duration={str(float(item_out.item.duration))}".')

        """
        We will trim the affected clip items if needed and then
        we will insert the transition track item in between both
        so the 'item_out' clip is automatically shifted the
        amount of time needed to fit the transition item in
        between.

        The transition has different types (modes), and according
        to those modes we will modify the durations and time
        moments of the clips affected as I show below, considering
        this case `[ClipA] [Gap] [ClipB]`:

        - TRIM: `[ClipA_trimmed] [Transition] [ClipB_trimmed]`
        - FREEZE_TAIL: `[ClipA] [Transition] [ClipB_delayed]`
        - FREEZE_HEAD: `[ClipA] [Transition] [ClipB_delayed]`
        - FREEZE_BOTH: `[ClipA] [Transition] [ClipB_delayed]`

        As you can see, only with the 'TRIM' mode we have to trim
        the clips, and with all the other modes we only have to
        delay (shift) the 'item_out', which is the normal behaviour
        when inserting a new item after the 'item_in' and before
        the 'item_out'.
        """
        # TODO: Do it...
        if mode in [TransitionMode.TRIM]:
            # Trim the end of 'item_in'
            self.shift_item_end_by(item_in.item, -duration)

            # Trim the start of 'item_out'
            """
            As we will use the first part of the 'item_out' clip
            for the transition, we need to adjust the media of
            that clip from the 't_start', but also the item itself
            from the 't_end' because we are trimming it in the
            track
            """
            self.shift_item_media_start_and_item_end_by(item_out.item, duration)

        # TODO: I need a way to read the lasts frames of a video
        # to obtain the ones from the left item for the transition
            
        print('pre adding transition item')

        # TODO: The 'TransitionTrackItem' needs to be like the
        # others

        """
        We use a time moment in between the 't_start' and the
        't_end' to make sure that it fits in the middle of the
        previous item so it is processed as it should be, using
        the 't_end' of the first element to be inserted there
        """
        t_start = (item_in.item.t_start + item_in.item.t_end) / 2

        transition_item: TransitionTrackItem = TransitionTrackItem(
            track = self._track,
            t_start = t_start,
            duration = duration,
            # TODO: The 'type' should be refactored
            type = type,
            mode = mode,
            # The 'item_in' and 'item_out' will be automatically set
            # according to the items that are previous and after the
            # transition time moment, and have been checked
            # previously. Including it here will make an infinite
            # loop as the next item would be referencing the
            # transition
            item_in = None,
            item_out = None
        )

        self.add_item(
            item = transition_item,
            do_force_add = True
        )

        return self
    
    def remove_transition(
        self,
        transition_item: 'TransitionTrackItem'
    ) -> '_TrackItems':
        """
        Remove the transition provided as `transition_item` from
        the track and readjust the items affected by that 
        transition. This means that if the transition was trimming
        the clips to be built, they will be enlarged again.
        """
        item_in = transition_item.item_in
        item_out = transition_item.item_out
        duration = transition_item.duration

        # Undone the changes the transition made to the
        # affected clips, if needed
        if transition_item.mode in [TransitionMode.TRIM]:
            # Enlarge the end of 'item_in'
            self.shift_item_end_by(item_in, duration)

            # Enlarge the start of 'item_out'
            """
            As we were using the first part of the 'item_out' clip
            for the transition, we need to adjust the media of 
            that clip from the 't_start', but also the item itself
            from the 't_end' because we trimmed it in the track
            """
            self.shift_item_media_start_and_item_end_by(item_out, -duration)

        self.remove_item(
            item = transition_item,
            do_create_gap_instead = False
        )

        # TODO: Do I really need this here (?)
        self._recalculate_all()

        return self
        

    """
    Methods related to shifting items
    """

    def shift_item_by(
        self,
        item: TrackItemType,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItems':
        """
        Shift the `item` provided the `delta` amount of time
        given.
        """
        if delta == 0:
            return self
        
        # TODO: Validate 'item'
        
        item.shift_by(delta)

        self._shift_next_elements(
            item = item,
            delta = delta,
            do_include_item = False
        )

        # TODO: Is this needed (?)
        self._recalculate_all()
        
        return self
    
    def shift_item_end_by(
        self,
        item: TrackItemType,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItems':
        """
        Shift the end of the `item` provided the `delta` amount
        of time given.
        """
        if delta == 0:
            return self
        
        # TODO: Validate 'item'
        
        item.shift_end_by(delta)

        self._shift_next_elements(
            item = item,
            delta = delta,
            do_include_item = False
        )

        # TODO: Is this needed (?)
        self._recalculate_all()
        
        return self
    
    def shift_item_media_start_and_item_end_by(
        self,
        item: TrackItemType,
        delta: Union[int, float, Fraction]
    ) -> '_TrackItems':
        """
        Shift the start of the media associated to the `item`
        provided the `delta` amount of time given and the end
        of that `item` also that same amount of time.
        """
        if delta == 0:
            return self
        
        # TODO: Validate 'item'
        
        item.shift_media_start_and_item_end_by(delta)

        self._shift_next_elements(
            item = item,
            delta = delta,
            do_include_item = False
        )

        # TODO: Is this needed (?)
        self._recalculate_all()
        
        return self

class _Track(ABC):
    """
    TODO: This class should replace the old one.

    Abstract class to be inherited by the different
    track item classes we implement.
    """

    @property
    def t_end(
        self
    ) -> Fraction:
        """
        The maximum `t_end` time moment, which is the biggest `t_end`
        value of a non-gap track item.
        """
        for item in reversed(self.items.items):
            if not is_gap_track_item(item):
                return item.t_end
            
        return 0

    @property
    def fps(
        self
    ) -> float:
        """
        The fps of the track, which are the same than the timeline
        that includes this track.
        """
        return self._timeline.fps
    
    @property
    def audio_fps(
        self
    ) -> float:
        """
        The audio fps of the track, which are the same than the
        timeline that includes this track.
        """
        return self._timeline.audio_fps
    
    @property
    def is_muted(
        self
    ) -> bool:
        """
        Flag to indicate if the track is muted or
        not. Being muted means that no audio frames
        will be retured from this track.
        """
        return self._is_muted
    
    @property
    def _t_utils(
        self
    ) -> '_TUtils':
        """
        Shortcut to the timeline `_TUtils` instance to work
        with time moments according to the timeline fps.
        """
        return self._timeline._t_utils
    
    def __init__(
        self,
        timeline: 'Timeline',
        index: int,
    ):
        self._timeline: 'Timeline' = timeline
        """
        The timeline instance this track belongs to.
        """
        # TODO: Initialize with a 'GapTrackItem' that lasts the whole track instead
        self.items = _TrackItems(self)
        """
        The list of items this track contains, that will be
        always ordered according to the `t_start` time moment.
        """

        self.index: int = index
        """
        The index of the track within the timeline.
        """
        self._is_muted: bool = False
        """
        Internal flag to indicate if the track is
        muted or not.
        """

    def mute(
        self
    ) -> '_Track':
        """
        Set the track as muted so no audio frame will
        be played from this track.
        """
        self._is_muted = True

        return self

    def unmute(
        self
    ) -> '_Track':
        """
        Set the track as unmuted so the audio frames
        will be played as normal.
        """
        self._is_muted = False

        return self

    """
    Functionality related to the items below
    """
    def is_gap_at(
        self,
        t_start: Union[int, float, Fraction],
        t_end: Union[int, float, Fraction, None] = None
    ) -> bool:
        """
        Check if the if the `t_start` time moment provided is
        occupied by a `GapTrackItem` instance, if only the
        `t_start` parameter is provided, or if the whole time
        interval from `t_start` to `t_end` is occupied by the
        same `GapTrackItem` instance.
        """
        item_at_t = self.get_item_at(t_start).item
        is_gap = PythonValidator.is_instance_of(item_at_t, GapTrackItem)

        return (
            is_gap
            if t_end is None else
            (
                is_gap and
                item_at_t.t_start >= t_start and
                # TODO: Should this be '<=' or only '<' (?)
                item_at_t.t_end <= t_end
            )
        )

    def get_item_at(
        self,
        t: Union[int, float, Fraction]
    ) -> '_TrackItemFound':
        """
        Get the item that is placed at the given `t` global
        time moment in this track.

        We will have always at least one item at that position
        because we fill empty gaps with the special gap item.
        """
        return self.items.get_item_at(t)
    
    def add_video_item(
        self,
        media: 'VideoMedia',
        t: Union[int, float, Fraction],
        do_force_add: bool = True
    ) -> '_Track':
        """
        Create a video track item and set it in the `t` global
        time moment provided.
        """
        from yta_editor.tracks.items.abstract import TrackItemWithVideoMedia

        # TODO: Validate 'media'
        
        return self.items.add_item(
            item = TrackItemWithVideoMedia(
                track = self,
                t_start = t,
                media = media
            ),
            do_force_add = do_force_add
        )
    
    def remove_item(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem']
    ) -> '_Track':
        """
        Remove the item from the list. The item, if existing
        and different to a gap, will be transformed into a
        gap.
        
        The list will have one item less but will keep being
        ordered by the t_start time moment.
        """
        self.items.remove_item(item)

        return self
    
    def remove_item_at(
        self,
        t: Union[int, float, Fraction]
    ) -> '_Track':
        """
        Remove the item existing at the `t` time moment
        provided, that will turn it into a gap (if it is
        not already a gap).
        """
        self.items.remove_item_at(t)

        return self
    
    """
    Methods related to transitions below
    """
    def add_transition(
        self,
        # TODO: We need to know the way to handle this 'type'
        # and maybe all the properties of the transition
        item_in: TrackItemType,
        item_out: TrackItemType,
        type: any = 'random',
        mode: TransitionMode = TransitionMode.TRIM,
        duration: Union[int, float, Fraction] = 1
    ) -> '_Track':
        """
        Create a transition between the `item_in` and `item_out`
        provided of the given type, if possible, and place
        it as a new item in between those 2 items that have
        to be consecutive and valid.
        """
        self.items.add_transition(
            item_in = item_in,
            item_out = item_out,
            type = type,
            mode = mode,
            duration = duration
        )

        return self
    
    def remove_transition(
        self,
        transition_item: 'TransitionTrackItem'
    ) -> '_TrackItems':
        """
        Remove the transition provided as `transition_item` from
        the track and readjust the items affected by that 
        transition. This means that if the transition was trimming
        the clips to be built, they will be enlarged again.
        """
        self.items.remove_transition(
            transition_item = transition_item
        )

        return self
    
    """
    Here below you have all the methods related to shifting
    and moving the items.

    TODO: Maybe we should move these methods to the 
    TrackItems class instead.
    """
    def shift_item_by(
        self,
        item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem'],
        delta: Union[int, float, Fraction]
    ) -> '_Track':
        """
        Remove the `item`, shift it the `delta` amount of time
        provided, and place it again in the track.
        """
        # TODO: Maybe valdiate some limits
        self.remove_item(item)
        item._shift_by(delta)
        self.items.add_item(item, True)

        return item
    
    def __str__(
        self
    ) -> str:
        """
        Stringify the instance by showing all the items.
        """
        return f'  - -   T   R   A   C   K   - - \n' + '\n'.join([
            f'[{str(int(index + 1))}] {item.__str__()}'
            for index, item in enumerate(self.items.items)
        ])
        
    
# TODO: This class has been created to wrap the methods
# to add transitions to the clips of a track due to its
# complexity. Maybe we will delete it as we are not
# handling the transitions as videos overlapping
class _TrackTransitionHandler:
    """
    *For internal use only*

    Class to handle the way we add transitions in between
    clips in a track.
    """

    @staticmethod
    def get_adjacent_medias_at(
        clips: list['Clip'],
        t: float
    ) -> tuple[Union['Clip', None], Union['Clip', None]]:
        """
        Get the 2 clips that are adjacent at the `t` time
        moment provided according to the given list of
        `clips`, or a tuple of None if there are no adjacent
        clips.
        """
        time_intervals = [
            (clip.t_start, clip.t_end)
            for clip in clips
        ]

        previous_index, next_index = _TrackTransitionHandler.get_adjacent_indexes_at(
            time_intervals = time_intervals,
            t = t
        )

        if (
            previous_index is None or
            next_index is None
        ):
            raise Exception(f'There are no adjacent medias at t={str(float(t))}')

        return (
            clips[previous_index],
            clips[next_index]
        )
    
    # TODO: This maybe should be somewhere else maybe
    def get_adjacent_indexes_at(
        time_intervals: list[tuple[float, float]],
        t: float
    ) -> Union[tuple[int, int], tuple[None, None]]:
        """
        Get the pair of indexes of the provided `time_invervals`
        list that are adjacent at the given `t` time moment (that
        must be exactly an `t_end` time moment).

        Returns
        -------
        (int, int) | (None, None)
            Indices of the adjacent ranges in the original list, or
            (None, None) if not found.
        """
        # Numbers are converted to float for comparison robustness
        t = float(t)
        # Keep the original index with the tuples
        indexed_ranges = list(enumerate(time_intervals))
        # Order by 't_start' without losing the original index
        sorted_ranges = sorted(indexed_ranges, key = lambda time_interval: float(time_interval[1][0]))

        for i in range(len(sorted_ranges) - 1):
            (idx_a, (_, end_a)) = sorted_ranges[i]
            (idx_b, (start_b, _)) = sorted_ranges[i + 1]

            previous_end_time = float(end_a)
            next_start_time = float(start_b)

            if _TrackTransitionHandler.are_adjacent_at(
                first_clip_end = previous_end_time,
                second_clip_start = next_start_time,
                t = t
            ):
                return (
                    idx_a,
                    idx_b
                )

        return (
            None,
            None
        )
    
    def are_adjacent_at(
        first_clip_end: float,
        second_clip_start: float,
        t: float
    ) -> bool:
        """
        Check if the `first_clip_end` and `second_clip_start`
        time moments provided are adjacent at the `t` time
        moment also given.
        """
        return (
            check_values_are_same(first_clip_end, second_clip_start)
            and check_values_are_same(t, first_clip_end)
        )

    @staticmethod
    def find_clips_overlapping_at_t(
        clips: list['Clip'],
        t: float
    ) -> tuple[Union['Clip', None], Union['Clip', None]]:
        """
        Find the list of 2 clips of the given `clips` list
        that are overlapping at the `t` time moment provided.

        This method will return a tuple including the
        previous and the next clips (that can be None if
        they are not overlapping).

        TODO: This will not happen because we don't accept
        overlapping clips in the same track. Remove it in the
        next commit.
        """
        prev_clip = None
        next_clip = None

        for clip in clips:
            if clip.t_end <= t:
                prev_clip = clip
            elif clip.t_start >= t:
                next_clip = clip
                break

        return (
            prev_clip,
            next_clip
        )
    
    @staticmethod
    def find_time_clips_overlap(
        previous_clip: 'Clip',
        next_clip: 'Clip'
    ) -> Union[tuple[float, float, float], None]:
        """
        Get the time interval in between the `previous_clip`
        and the `next_clip` provided are colliding, getting
        None if they are not colliding, or a tuple including:
        - `overlap_start`
        - `overlap_end`
        - `overlap_time`

        TODO: This will not happen because we don't accept
        overlapping clips in the same track. Remove it in the
        next commit.
        """
        # Overlapping time range
        overlap_start = max(previous_clip.t_start, next_clip.t_start)
        overlap_end = min(previous_clip.t_end, next_clip.t_end)

        if overlap_start >= overlap_end:
            # No overlap
            return None
            print("Advertencia: no hay solapamiento real, creando transición forzada.")
            overlap_start = t_start
            overlap_end = t_start + duration

        return (
            overlap_start,
            overlap_end,
            overlap_end - overlap_start
        )
    
class _TrackWithAudio(_Track):
    """
    Class that has the ability to obtain the
    audio frames from the source and must be
    inherited by those tracks that have audio
    (or video, that includes audio).
    """

    def __init__(
        self,
        # TODO: Where does it come from (?)
        audio_samples_per_frame: int,
        audio_layout: str,
        audio_format: str,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.audio_samples_per_frame: int = audio_samples_per_frame
        """
        The number of samples per audio frame.
        """
        self.audio_layout: str = audio_layout
        """
        The layout of the audio, that can be 'mono'
        or 'stereo'.
        """
        self.audio_format: str = audio_format
        """
        The format of the audio, that can be 's16',
        'flt', 'fltp', etc.
        """

    # TODO: This is not working well when
    # source has different fps
    def get_audio_frames_at(
        self,
        t: Union[int, float, Fraction],
        # TODO: Ignore this, we have the 'self.fps'
        video_fps: Union[int, float, Fraction] = None,
    ):
        """
        Get the sequence of audio frames that
        must be displayed at the 't' time 
        moment provided, which the collection
        of audio frames corresponding to the
        audio that is being played at that time
        moment.

        Remember, this 't' time moment provided
        is about the track, and we make the
        conversion to the actual audio 't' to
        get the frame.
        """
        frames = (
            generate_silent_frames(
                video_fps = self.fps,
                audio_fps = self.audio_fps,
                audio_samples_per_frame = self.audio_samples_per_frame,
                layout = self.audio_layout,
                format = self.audio_format
            )
            if self.is_muted else
            # TODO: What if we find None here (?)
            self.get_item_at(t).item.get_audio_frames_at(t, self.fps)
        )

        for frame in frames:
            print('      HELLO       ')
            print(frame)
            if frame == []:
                print(f'   [ERROR] Audio frame t:{str(float(t))} not obtained.')
            yield frame

class _TrackWithVideo(_Track):
    """
    Class that has the ability to obtain the
    video frames from the source and must be
    inherited by those tracks that have video.
    """

    def __init__(
        self,
        size: tuple[int, int],
        **kwargs
    ):
        super().__init__(**kwargs)

        # TODO: This is not needed actually...
        self.size: tuple[int, int] = size
        """
        The size that the output video will have.
        """
        self._transitions: list['TransitionTrackItem'] = []
        """
        List of transitions that exist in the track and are
        joining adjacent videos.
        """

    def get_video_frame_at(
        self,
        t_track: Union[int, float, Fraction]
    ) -> 'VideoFrameWrapped':
        """
        Get the frame that must be displayed at the
        `t_track` global time moment provided, which
        is a frame from the video that is being
        played at that time moment.

        The global `t_track` time moment provided will
        be transformed into the local `t_track` time
        moment for the specific track item located at
        that position to obtain the frame.
        """
        # TODO: What if the frame, that comes from
        # a video, doesn't have the expected size (?)
        return self.get_item_at(t_track).item._get_video_frame_at(t_track)

def is_gap_track_item(
    item: Union['_AudioTrackItem', '_VideoTrackItem', 'TransitionTrackItem', 'GapTrackItem']
) -> bool:
    """
    Check if the `item` provided is an instance of a
    `GapTrackItem` or not.
    """
    return PythonValidator.is_instance_of(item, 'GapTrackItem')
    
"""
Note for the developers:
- When we are working with media and a track we need
to make sure that the time moment we are using matches
the fps of that track and it is a multiple of 1/fps to
be able to position that media properly.
"""